import unittest
from datetime import datetime, timedelta

import numpy as np

import finam as fm


class TestNoise(unittest.TestCase):
    def test_noise_scalar_0d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.NoGrid(),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect()
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"][0, ...]
        self.assertEqual(data.shape, ())

    def test_noise_uniform_1d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20,)),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"][0, ...]
        self.assertEqual(data.shape, (19,))

    def test_noise_uniform_2d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15)),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"][0, ...]
        self.assertEqual(data.shape, (19, 14))

    def test_noise_uniform_3d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UniformGrid((20, 15, 10)),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"][0, ...]
        self.assertEqual(data.shape, (19, 14, 9))

    def test_noise_points_1d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UnstructuredPoints(np.random.random((100, 1))),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"][0, ...]
        self.assertEqual(data.shape, (100,))

    def test_noise_points_2d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UnstructuredPoints(np.random.random((100, 2))),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"][0, ...]
        self.assertEqual(data.shape, (100,))

    def test_noise_points_3d(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.UnstructuredPoints(np.random.random((100, 3))),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        composition.connect(time)
        composition.run(end_time=datetime(2000, 1, 2))

        data = sink.data["Input"][0, ...]
        self.assertEqual(data.shape, (100,))

    def test_noise_fail_nogrid(self):
        time = datetime(2000, 1, 1)
        in_info = fm.Info(
            time=None,
            grid=fm.NoGrid(1),
            units="m",
        )

        source = fm.components.SimplexNoise(info=in_info)
        trigger = fm.components.TimeTrigger(
            in_info=fm.Info(time=None, grid=None, units=None),
            start=time,
            step=timedelta(days=1),
        )
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = fm.Composition([source, trigger, sink])

        source.outputs["Noise"] >> trigger.inputs["In"]
        trigger.outputs["Out"] >> sink.inputs["Input"]

        with self.assertRaises(fm.FinamMetaDataError):
            composition.connect(time)

    def test_noise_fail(self):
        with self.assertRaises(ValueError):
            _source = fm.components.SimplexNoise(octaves=0)

        source = fm.components.SimplexNoise(octaves=1)
        source._update()


class TestStaticNoise(unittest.TestCase):
    def test_static_noise(self):
        in_info = fm.Info(
            time=None,
            grid=fm.NoGrid(),
            units="m",
        )

        source = fm.components.StaticSimplexNoise(info=in_info, seed=123)
        """
        sink = fm.components.DebugConsumer(
            {"Input": fm.Info(None, grid=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )
        """
        # We want to get to a point where this works
        sink = fm.components.DebugPushConsumer(
            inputs={
                "Input": fm.Info(time=None, grid=fm.NoGrid(), units="m"),
            },
            log_data="INFO",
        )

        composition = fm.Composition([source, sink])

        source.outputs["Noise"] >> sink.inputs["Input"]

        composition.connect(None)
        data_1 = sink.data["Input"][0, ...]

        self.assertEqual(data_1.shape, ())

        composition.run(end_time=None)

        data_2 = sink.data["Input"][0, ...]
        self.assertEqual(data_1, data_2)
        self.assertEqual(data_2.shape, ())


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timedelta

import finam as fm


class TestScheduleLogger(unittest.TestCase):
    def test_schedule_logger(self):
        start = datetime(2000, 1, 1)
        info = fm.Info(time=start, grid=fm.NoGrid())

        module1 = fm.components.CallbackGenerator(
            callbacks={"Out": (lambda t: t.day, info)},
            start=start,
            step=timedelta(days=2),
        )
        module2 = fm.components.CallbackComponent(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda inp, _t: {"Out": inp["In"][0, ...]},
            start=start,
            step=timedelta(days=5),
        )
        module3 = fm.components.CallbackComponent(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Out": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda inp, _t: {"Out": inp["In"][0, ...]},
            start=start,
            step=timedelta(days=8),
        )

        schedule = fm.components.ScheduleLogger(
            inputs={
                "M1": True,
                "M2": True,
                "M3": True,
            },
            log_level="DEBUG",
            stdout=True,
        )

        composition = fm.Composition([module1, module2, module3, schedule])

        module1.outputs["Out"] >> fm.adapters.Scale(1.0) >> module2.inputs["In"]
        module2.outputs["Out"] >> fm.adapters.Scale(1.0) >> module3.inputs["In"]

        module1.outputs["Out"] >> schedule.inputs["M1"]
        module2.outputs["Out"] >> schedule.inputs["M2"]
        module3.outputs["Out"] >> schedule.inputs["M3"]

        composition.connect(start)
        composition.run(end_time=datetime(2000, 1, 30))


class TestPushDebugConsumer(unittest.TestCase):
    def test_consumer(self):
        start = datetime(2000, 1, 1)
        info = fm.Info(time=start, grid=fm.NoGrid())

        module1 = fm.components.CallbackGenerator(
            callbacks={"Out": (lambda t: t.day, info)},
            start=start,
            step=timedelta(days=2),
        )
        consumer = fm.components.DebugPushConsumer(
            inputs={
                "In": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callbacks={
                "In": lambda n, d, t: print(t),
            },
            log_data="INFO",
        )

        composition = fm.Composition([module1, consumer])

        module1.outputs["Out"] >> consumer.inputs["In"]

        composition.connect(start)
        self.assertEqual(consumer.data["In"][0, ...], 1)

        composition.run(start_time=start, end_time=datetime(2000, 1, 10))

        self.assertEqual(consumer.data["In"][0, ...], 11)


if __name__ == "__main__":
    unittest.main()



import logging
import sys
import unittest

from finam.tools.log_helper import ErrorLogger, LogCStdOutStdErr, LogStdOutStdErr
from finam.tools.wurlitzer import libc


def raise_and_log(do_log):
    with ErrorLogger(logging.getLogger(None), do_log=do_log):
        raise ValueError("This is an Error!")


class TestLog(unittest.TestCase):
    def test_error_log(self):
        with self.assertLogs() as captured:
            with self.assertRaises(ValueError):
                raise_and_log(do_log=True)
        self.assertEqual(len(captured.records), 1)
        self.assertEqual(captured.records[0].levelno, logging.ERROR)
        # suppress logging
        with self.assertLogs(level=logging.WARN) as captured:
            with self.assertRaises(ValueError):
                raise_and_log(do_log=False)
            # "assertNoLogs" is only available in python >=3.10
            # so we need to at least log something to prove,
            # nothing else was logged
            logging.getLogger(None).warning("Dummy warning.")
        # check that we only got the one dummy log
        self.assertEqual(len(captured.records), 1)

    def test_log_levels(self):
        with self.assertLogs("foo", level="TRACE") as captured:
            logging.getLogger("foo").trace("A")
            logging.getLogger("foo").profile("B")

        self.assertEqual(len(captured.records), 2)
        self.assertEqual(captured.records[0].levelno, logging.TRACE)
        self.assertEqual(captured.records[0].message, "A")
        self.assertEqual(captured.records[1].levelno, logging.PROFILE)
        self.assertEqual(captured.records[1].message, "B")

    def test_redirect(self):
        with self.assertLogs() as captured:
            with LogStdOutStdErr():
                print("Hi from Python")
                print("Boo from Python", file=sys.stderr)
        self.assertEqual(len(captured.records), 2)
        self.assertEqual(captured.records[0].levelno, logging.INFO)
        self.assertEqual(captured.records[0].message, "Hi from Python")
        self.assertEqual(captured.records[1].levelno, logging.WARNING)
        self.assertEqual(captured.records[1].message, "Boo from Python")

    def test_c_redirect(self):
        with self.assertLogs() as captured:
            with LogCStdOutStdErr():
                libc.puts(b"Hi from C")
        self.assertEqual(len(captured.records), 1)
        self.assertEqual(captured.records[0].levelno, logging.INFO)
        self.assertEqual(captured.records[0].message, "Hi from C")


if __name__ == "__main__":
    unittest.main()



"""Generic component based on a callback."""

from datetime import datetime

from ..sdk import TimeComponent
from ..tools.date_helper import is_timedelta
from ..tools.log_helper import ErrorLogger

__all__ = ["CallbackComponent"]


class CallbackComponent(TimeComponent):
    """Component to generate, transform or consume data in fixed time intervals using a callback.

    .. code-block:: text

                   +-------------------+
      --> [custom] |                   | [custom] -->
      --> [custom] | CallbackComponent |
      --> [......] |                   | [......] -->
                   +-------------------+

    Examples
    --------

    .. testcode:: constructor

        import datetime as dt
        import finam as fm

        component = fm.components.CallbackComponent(
            inputs={
                "A": fm.Info(time=None, grid=fm.NoGrid()),
                "B": fm.Info(time=None, grid=fm.NoGrid()),
            },
            outputs={
                "Sum": fm.Info(time=None, grid=fm.NoGrid()),
                "Diff": fm.Info(time=None, grid=fm.NoGrid()),
            },
            callback=lambda inp, _t: {
                "Sum": inp["A"] + inp["B"],
                "Diff": inp["A"] - inp["B"],
            },
            start=dt.datetime(2000, 1, 1),
            step=dt.timedelta(days=7),
        )

    .. testcode:: constructor
        :hide:

        component.initialize()

    Parameters
    ----------
    inputs : dict[str, Info]
        Input names and data info.
    outputs : dict[name, Info]
        Output names and data info.
    callback
        Callback f({inputs}, time) -> {outputs}
    start : :class:`datetime <datetime.datetime>`
        Start date and time
    step : :class:`timedelta <datetime.timedelta>` or :class:`relativedelta <dateutil.relativedelta.relativedelta>`
        Time step
    initial_pull : bool, optional
        whether to pull initial data. The first call of the callback with have ``None`` for inputs of disabled.
        Default ``True``.
    """

    def __init__(self, inputs, outputs, callback, start, step, initial_pull=True):
        super().__init__()

        with ErrorLogger(self.logger):
            if not isinstance(start, datetime):
                raise ValueError("Start must be of type datetime")
            if not is_timedelta(step):
                raise ValueError("Step must be of type timedelta or relativedelta")

        self._input_infos = inputs
        self._output_infos = outputs
        self._callback = callback
        self._step = step
        self._time = start
        self._initial_pull = initial_pull
        self._data_generated = False

    def _next_time(self):
        return self.time + self._step

    def _initialize(self):
        for name, info in self._input_infos.items():
            info.time = self.time
            self.inputs.add(name=name, info=info)

        for name, info in self._output_infos.items():
            info.time = self.time
            self.outputs.add(name=name, info=info)

        pull_data = list(self._input_infos) if self._initial_pull else {}

        self.create_connector(pull_data=pull_data)

    def _connect(self, start_time):
        push_data = {}
        if not self._data_generated:
            if self._initial_pull:
                if self.connector.all_data_pulled:
                    push_data = self._callback(self.connector.in_data, self.time)
                    self._data_generated = True
            else:
                push_data = self._callback(None, self.time)
                self._data_generated = True

        self.try_connect(start_time, push_data=push_data)

    def _validate(self):
        pass

    def _update(self):
        self._time += self._step

        inp = {n: self.inputs[n].pull_data(self.time) for n in self._input_infos.keys()}
        outp = self._callback(inp, self.time)
        for name, val in outp.items():
            if val is not None:
                self.outputs[name].push_data(val, self.time)

    def _finalize(self):
        pass



"""Capture C-level FD output on pipes

Use `wurlitzer.pipes` or `wurlitzer.sys_pipes` as context managers.

This file was copied from: https://github.com/minrk/wurlitzer

In the specific version of:
https://github.com/minrk/wurlitzer/blob/1ad008a1e163356475696b3693092c27ad4f710a/wurlitzer.py

wurlitzer is released under the MIT license and was copied here as is.
"""
from __future__ import print_function

__version__ = "3.0.3.dev"

__all__ = [
    "pipes",
    "sys_pipes",
    "sys_pipes_forever",
    "stop_sys_pipes",
    "Wurlitzer",
]

import ctypes
import errno
import io
import os
import platform
import selectors
import sys
import threading
import time
import warnings
from contextlib import contextmanager
from functools import lru_cache, partial
from queue import Empty, Queue

if os.name != "nt":
    from fcntl import F_GETFL, F_SETFL, fcntl

    try:
        from fcntl import F_GETPIPE_SZ, F_SETPIPE_SZ
    except ImportError:
        # ref: linux uapi/linux/fcntl.h
        F_SETPIPE_SZ = 1024 + 7
        F_GETPIPE_SZ = 1024 + 8


# Windows support adapted from
# https://github.com/chrisjbillington/zprocess@89b464c3a2d10cb3282bea625670f807430adf03
# MIT License


def _get_streams_windows():
    """Get file pointer for C stream"""

    class FILE(ctypes.Structure):
        """FILE struct for Windows"""

        _fields_ = [
            ("_ptr", ctypes.c_char_p),
            ("_cnt", ctypes.c_int),
            ("_base", ctypes.c_char_p),
            ("_flag", ctypes.c_int),
            ("_file", ctypes.c_int),
            ("_charbuf", ctypes.c_int),
            ("_bufsize", ctypes.c_int),
            ("_tmpfname", ctypes.c_char_p),
        ]

    iob_func = libc.__iob_func
    iob_func.restype = ctypes.POINTER(FILE)
    iob_func.argtypes = []
    streams = iob_func()
    return (ctypes.c_void_p(ctypes.addressof(streams[i])) for i in (1, 2))


def _get_streams_cffi():
    """Use CFFI to lookup stdout/stderr pointers

    Should work ~everywhere, but requires compilation
    """
    try:
        import cffi
    except ImportError:
        raise ImportError(
            "Failed to lookup stdout symbols in libc. Fallback requires cffi."
        )

    try:
        _ffi = cffi.FFI()
        _ffi.cdef("const size_t c_stdout_p();")
        _ffi.cdef("const size_t c_stderr_p();")
        _lib = _ffi.verify(
            "\n".join(
                [
                    "#include <stdio.h>",
                    "const size_t c_stdout_p() { return (size_t) (void*) stdout; }",
                    "const size_t c_stderr_p() { return (size_t) (void*) stderr; }",
                ]
            )
        )
        c_stdout_p = ctypes.c_void_p(_lib.c_stdout_p())
        c_stderr_p = ctypes.c_void_p(_lib.c_stderr_p())
    except Exception as e:
        warnings.warn(
            "Failed to lookup stdout with cffi: {}.\nStreams may not be flushed.".format(
                e
            )
        )
        return (None, None)
    else:
        return c_stdout_p, c_stderr_p


c_stdout_p = c_stderr_p = None
if os.name == "nt":
    libc = ctypes.cdll.msvcrt
    c_stdout_p, c_stderr_p = _get_streams_windows()
else:
    libc = ctypes.CDLL(None)
    try:
        c_stdout_p = ctypes.c_void_p.in_dll(libc, "stdout")
        c_stderr_p = ctypes.c_void_p.in_dll(libc, "stderr")
    except ValueError:
        # libc.stdout has a funny name on macOS
        try:
            c_stdout_p = ctypes.c_void_p.in_dll(libc, "__stdoutp")
            c_stderr_p = ctypes.c_void_p.in_dll(libc, "__stderrp")
        except ValueError:
            c_stdout_p, c_stderr_p = _get_streams_cffi()

libc.setvbuf.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_size_t,
]

STDOUT = 2
PIPE = 3

_default_encoding = getattr(sys.stdin, "encoding", None) or "utf8"
if _default_encoding.lower() == "ascii":
    # don't respect ascii
    _default_encoding = "utf8"  # pragma: no cover


def dup2(a, b, timeout=3):
    """Like os.dup2, but retry on EBUSY"""
    dup_err = None
    # give FDs 3 seconds to not be busy anymore
    for i in range(int(10 * timeout)):
        try:
            return os.dup2(a, b)
        except OSError as e:
            dup_err = e
            if e.errno == errno.EBUSY:
                time.sleep(0.1)
            else:
                raise
    if dup_err:
        raise dup_err


def _unbuffer(c_stream_p):
    """Set C output streams to unbuffered

    From chrisjbillington/zprocess (MIT License)
    """
    if os.name == "nt":
        _IONBF = 4
    else:
        _IONBF = 2
    libc.setvbuf(c_stream_p, None, _IONBF, 0)


def _nonblocking(fd):
    # Windows impl adapted from https://stackoverflow.com/questions/34504970/non-blocking-read-on-os-pipe-on-windows
    if os.name == "nt":
        import msvcrt
        from ctypes import POINTER, WinError, byref, windll
        from ctypes.wintypes import BOOL, DWORD, HANDLE

        LPDWORD = POINTER(DWORD)

        PIPE_NOWAIT = DWORD(1)
        SetNamedPipeHandleState = windll.kernel32.SetNamedPipeHandleState
        SetNamedPipeHandleState.argtypes = [HANDLE, LPDWORD, LPDWORD, LPDWORD]
        SetNamedPipeHandleState.restype = BOOL

        h = msvcrt.get_osfhandle(fd)

        res = windll.kernel32.SetNamedPipeHandleState(h, byref(PIPE_NOWAIT), None, None)
        if res == 0:
            print(WinError())
    else:
        flags = fcntl(fd, F_GETFL)
        fcntl(fd, F_SETFL, flags | os.O_NONBLOCK)


_WIN_ERROR_NO_DATA = 232


@lru_cache()
def _get_max_pipe_size():
    """Get max pipe size

    Reads /proc/sys/fs/pipe-max-size on Linux.
    Always returns None elsewhere.

    Returns integer (up to 1MB),
    or None if no value can be determined.

    Adapted from wal-e, (c) 2018, WAL-E Contributors
    used under BSD-3-clause
    """
    if platform.system() != "Linux":
        return

    # If Linux procfs (or something that looks like it) exposes its
    # maximum F_SETPIPE_SZ, adjust the default buffer sizes.
    try:
        with open("/proc/sys/fs/pipe-max-size", "r") as f:
            # Figure out OS max pipe size
            pipe_max_size = int(f.read())
    except Exception:
        pass
    else:
        if pipe_max_size > 1024 * 1024:
            # avoid unusually large values, limit to 1MB
            return 1024 * 1024
        elif pipe_max_size <= 65536:
            # smaller than default, don't do anything
            return None
        else:
            return pipe_max_size


class Wurlitzer:
    """Class for Capturing Process-level FD output via dup2

    Typically used via `wurlitzer.pipes`
    """

    flush_interval = 0.2

    def __init__(
        self,
        stdout=None,
        stderr=None,
        encoding=_default_encoding,
        bufsize=_get_max_pipe_size(),
    ):
        """
        Parameters
        ----------
        stdout: stream or None
            The stream for forwarding stdout.
        stderr = stream or None
            The stream for forwarding stderr.
        encoding: str or None
            The encoding to use, if streams should be interpreted as text.
        bufsize: int or None
            Set pipe buffer size using fcntl F_SETPIPE_SZ (linux only)
            default: use /proc/sys/fs/pipe-max-size up to a max of 1MB
            if 0, will do nothing.
        """
        self._stdout = stdout
        if stderr == STDOUT:
            self._stderr = self._stdout
        else:
            self._stderr = stderr
        self.encoding = encoding
        if bufsize is None:
            bufsize = _get_max_pipe_size()
        self._bufsize = bufsize
        self._save_fds = {}
        self._real_fds = {}
        self._handlers = {}
        self._handlers["stderr"] = self._handle_stderr
        self._handlers["stdout"] = self._handle_stdout

    def _setup_pipe(self, name):
        real_fd = getattr(sys, "__%s__" % name).fileno()
        save_fd = os.dup(real_fd)
        self._save_fds[name] = save_fd
        c_stream_p = globals()["c_{}_p".format(name)]
        if c_stream_p is not None:
            _unbuffer(c_stream_p)

        pipe_out, pipe_in = os.pipe()
        # set max pipe buffer size (linux only)
        if os.name != "nt" and self._bufsize:
            try:
                fcntl(pipe_in, F_SETPIPE_SZ, self._bufsize)
            except OSError as error:
                warnings.warn(
                    "Failed to set pipe buffer size: " + str(error), RuntimeWarning
                )

        dup2(pipe_in, real_fd)
        os.close(pipe_in)
        self._real_fds[name] = real_fd

        # make pipe_out non-blocking
        _nonblocking(pipe_out)
        return pipe_out

    def _decode(self, data):
        """Decode data, if any

        Called before passing to stdout/stderr streams
        """
        if self.encoding:
            data = data.decode(self.encoding, "replace")
        return data

    def _handle_stdout(self, data):
        if self._stdout:
            self._stdout.write(self._decode(data))

    def _handle_stderr(self, data):
        if self._stderr:
            self._stderr.write(self._decode(data))

    def _setup_handle(self):
        """Setup handle for output, if any"""
        self.handle = (self._stdout, self._stderr)

    def _finish_handle(self):
        """Finish handle, if anything should be done when it's all wrapped up."""
        pass

    def _flush(self):
        """flush sys.stdout/err and low-level FDs"""
        if self._stdout and sys.stdout:
            sys.stdout.flush()
        if self._stderr and sys.stderr:
            sys.stderr.flush()

        if os.name == "nt":
            # In windows we flush all output streams
            # by calling flush on a NULL pointer
            libc.fflush(ctypes.c_void_p())
        else:
            if c_stdout_p is not None:
                libc.fflush(c_stdout_p)

            if c_stderr_p is not None:
                libc.fflush(c_stderr_p)

    def _set_thread_priority(self):
        """Set thread priority

        no-op except on Windows

        Adapted from zprocess (MIT License)
        """
        if os.name == "nt":
            w32 = ctypes.windll.kernel32
            THREAD_SET_INFORMATION = 0x20
            THREAD_PRIORITY_ABOVE_NORMAL = 1
            handle = w32.OpenThread(
                THREAD_SET_INFORMATION, False, threading.current_thread().ident
            )
            result = w32.SetThreadPriority(handle, THREAD_PRIORITY_ABOVE_NORMAL)
            w32.CloseHandle(handle)
            if not result:
                print(
                    "Failed to set priority of thread:",
                    w32.GetLastError(),
                    file=sys.__stderr__,
                )

    def __enter__(self):
        return self.start()

    def start(self):
        """Start capturing output

        Typically used via context manager
        """
        # flush anything out before starting
        self._flush()
        # setup handle
        self._setup_handle()
        self._control_r, self._control_w = os.pipe()

        # create pipe for stdout
        draining = False
        names = {}
        pipes = []
        if self._stdout:
            pipe = self._setup_pipe("stdout")
            pipes.append(pipe)
            names[pipe] = "stdout"
        if self._stderr:
            pipe = self._setup_pipe("stderr")
            pipes.append(pipe)
            names[pipe] = "stderr"

        # flush pipes in a background thread to avoid blocking
        # the reader thread when the buffer is full
        flush_queue = Queue()

        def flush_main():
            self._set_thread_priority()
            msg = ""
            while True:
                try:
                    msg = flush_queue.get(timeout=self.flush_interval)
                except Empty:
                    # flush every flush_interval,
                    # even if we get no input
                    pass
                self._flush()
                if msg == "stop":
                    return

        flush_thread = threading.Thread(target=flush_main)
        flush_thread.daemon = True
        flush_thread.start()

        # run one thread per pipe
        # because Windows can only do blocking reads on pipes
        # Windows can't select on pipes,
        # and even ProactorEventLoop.connect_read_pipe doesn't actually support pipes.

        def pipe_forwarder(fd, name, handler):
            """Forward bytes on a pipe to stream messages"""
            self._set_thread_priority()
            if os.name == "nt":
                # can't poll on Windows, sleep instead
                # this shouldn't come up because reads are always blocking
                def poll():
                    time.sleep(self.flush_interval)

            else:
                poller = selectors.DefaultSelector()
                poller.register(fd, selectors.EVENT_READ)

                def poll():
                    poller.select(self.flush_interval)

            while True:
                try:
                    data = os.read(fd, 1024)
                except OSError as e:
                    if os.name == "nt":
                        w32 = ctypes.windll.kernel32
                        no_data = (
                            e.errno == errno.EINVAL
                            and w32.GetLastError() == _WIN_ERROR_NO_DATA
                        )
                    else:
                        no_data = e.errno == errno.EAGAIN
                    if no_data:
                        if draining:
                            break
                        else:
                            poll()
                            continue
                    else:
                        raise
                if not data:
                    break
                else:
                    try:
                        handler(data)
                    except Exception as e:
                        # FIXME: this would produce an infinite loop on stderr
                        if name != "stderr":
                            print(
                                "Error handling pipe bytes: {}".format(e),
                                file=sys.__stderr__,
                            )
                        break
            if os.name != "nt":
                poller.close()
            # done reading, close pipe
            os.close(fd)

        def forwarder_control():
            self._set_thread_priority()
            nonlocal draining
            pipe_threads = []
            for fd in pipes:
                name = names[fd]
                handler = handler = getattr(self, "_handle_%s" % name)
                pipe_thread = threading.Thread(
                    target=pipe_forwarder, args=(fd, name, handler)
                )
                pipe_thread.daemon = True
                pipe_thread.start()
                pipe_threads.append(pipe_thread)

            msg = os.read(self._control_r, 1)
            os.close(self._control_r)

            # stop flush thread
            flush_queue.put("stop")
            flush_thread.join(timeout=10)
            draining = True

            # stop pipe threads
            for pipe_thread in pipe_threads:
                pipe_thread.join(timeout=10)
                if pipe_thread.is_alive():
                    print("Pipe still alive!", pipe_thread, file=sys.__stderr__)

        self.thread = threading.Thread(target=forwarder_control)
        self.thread.daemon = True
        self.thread.start()

        return self.handle

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def stop(self):
        """Stop capturing output"""
        # flush before exiting
        self._flush()

        # restore original state, close pipes
        # forwarder thread will finish when it reaches EOF

        # signal output is complete on control channel
        os.write(self._control_w, b"\1")
        os.close(self._control_w)

        # finally, wait for thread to finish
        self.thread.join(timeout=10)

        for name, real_fd in self._real_fds.items():
            save_fd = self._save_fds[name]
            dup2(save_fd, real_fd)
            os.close(save_fd)

        if self.thread.is_alive():
            print("Wurlitzer thread failed to complete!", file=sys.__stderr__)

        # finalize handle
        self._finish_handle()


@contextmanager
def pipes(stdout=PIPE, stderr=PIPE, encoding=_default_encoding, bufsize=None):
    """Capture C-level stdout/stderr in a context manager.

    The return value for the context manager is (stdout, stderr).

    .. versionchanged:: 3.0

        when using `PIPE` (default), the type of captured output
        is `io.StringIO/BytesIO` instead of an OS pipe.
        This eliminates max buffer size issues (and hang when output exceeds 65536 bytes),
        but also means the buffer cannot be read with `.read()` methods
        until after the context exits.

    Examples
    --------

    >>> with capture() as (stdout, stderr):
    ...     printf("C-level stdout")
    ... output = stdout.read()
    """
    stdout_pipe = stderr_pipe = False
    if encoding:
        PipeIO = partial(io.StringIO, newline=None)
    else:
        PipeIO = io.BytesIO
    # setup stdout
    if stdout == PIPE:
        stdout_r = stdout_w = PipeIO()
        stdout_pipe = True
    else:
        stdout_r = stdout_w = stdout
    # setup stderr
    if stderr == STDOUT:
        stderr_r = None
        stderr_w = stdout_w
    elif stderr == PIPE:
        stderr_r = stderr_w = PipeIO()
        stderr_pipe = True
    else:
        stderr_r = stderr_w = stderr
    w = Wurlitzer(stdout=stdout_w, stderr=stderr_w, encoding=encoding, bufsize=bufsize)
    try:
        with w:
            yield stdout_r, stderr_r
    finally:
        # close pipes
        if stdout_pipe:
            # seek to 0 so that it can be read after exit
            stdout_r.seek(0)
        if stderr_pipe:
            # seek to 0 so that it can be read after exit
            stderr_r.seek(0)


def sys_pipes(encoding=_default_encoding, bufsize=None):
    """Redirect C-level stdout/stderr to sys.stdout/stderr

    This is useful of sys.sdout/stderr are already being forwarded somewhere.

    DO NOT USE THIS if sys.stdout and sys.stderr are not already being forwarded.
    """
    return pipes(sys.stdout, sys.stderr, encoding=encoding, bufsize=bufsize)


_mighty_wurlitzer = None
_mighty_lock = threading.Lock()


def sys_pipes_forever(encoding=_default_encoding, bufsize=None):
    """Redirect all C output to sys.stdout/err

    This is not a context manager; it turns on C-forwarding permanently.
    """
    global _mighty_wurlitzer
    with _mighty_lock:
        if _mighty_wurlitzer is None:
            _mighty_wurlitzer = sys_pipes(encoding, bufsize)
            _mighty_wurlitzer.__enter__()


def stop_sys_pipes():
    """Stop permanent redirection started by sys_pipes_forever"""
    global _mighty_wurlitzer
    with _mighty_lock:
        if _mighty_wurlitzer is not None:
            _mighty_wurlitzer.__exit__(None, None, None)
            _mighty_wurlitzer = None


_extension_enabled = False


def load_ipython_extension(ip):
    """Register me as an IPython extension

    Captures all C output during execution and forwards to sys.

    Does nothing on terminal IPython.

    Use: %load_ext wurlitzer
    """
    global _extension_enabled

    if not getattr(ip, "kernel", None):
        warnings.warn("wurlitzer extension doesn't do anything in terminal IPython")
        return
    for name in ("__stdout__", "__stderr__"):
        if getattr(sys, name) is None:
            warnings.warn("sys.{} is None. Wurlitzer can't capture output without it.")
            return

    ip.events.register("pre_execute", sys_pipes_forever)
    ip.events.register("post_execute", stop_sys_pipes)
    _extension_enabled = True


def unload_ipython_extension(ip):
    """Unload me as an IPython extension

    Use: %unload_ext wurlitzer
    """
    global _extension_enabled
    if not _extension_enabled:
        return

    ip.events.unregister("pre_execute", sys_pipes_forever)
    ip.events.unregister("post_execute", stop_sys_pipes)
    # sys_pipes_forever was called in pre_execute
    # after unregister we need to call it explicitly:
    stop_sys_pipes()
    _extension_enabled = False



"""
Basic linear and nearest neighbour regridding adapters.

See package `finam-regrid <https://finam.pages.ufz.de/finam-regrid/>`_ for more advanced regridding.
"""

from abc import ABC, abstractmethod

import numpy as np
from pyproj import Transformer, crs
from scipy.interpolate import LinearNDInterpolator, RegularGridInterpolator
from scipy.spatial import KDTree

from ..data import tools as dtools
from ..data.grid_spec import StructuredGrid
from ..errors import FinamDataError, FinamMetaDataError
from ..sdk import Adapter
from ..tools.log_helper import ErrorLogger

__all__ = [
    "ARegridding",
    "RegridNearest",
    "RegridLinear",
]


class ARegridding(Adapter, ABC):
    """Abstract regridding class for handling data info"""

    def __init__(self, in_grid=None, out_grid=None, out_mask=None):
        super().__init__()
        self.input_grid = in_grid
        self.output_grid = out_grid
        if dtools.mask_specified(out_mask) and out_mask is not None:
            out_mask = np.ma.make_mask(out_mask, shrink=False)
        self.output_mask = out_mask
        self.downstream_mask = None
        self.input_mask = None
        self.input_meta = None
        self.transformer = None
        self._is_initialized = False
        self._out_mask_checked = False

    @abstractmethod
    def _update_grid_specs(self):
        """set up interpolator"""

    def _get_info(self, info):
        request = info.copy_with(grid=self.input_grid, mask=None)
        in_info = self.exchange_info(request)

        if self.output_grid is None and info.grid is None:
            with ErrorLogger(self.logger):
                raise FinamMetaDataError("Missing target grid specification")
        if self.input_grid is None and in_info.grid is None:
            with ErrorLogger(self.logger):
                raise FinamMetaDataError("Missing source grid specification")

        if self.output_mask is None and info.mask is None:
            with ErrorLogger(self.logger):
                raise FinamMetaDataError("Missing target mask specification")
        if self.input_mask is None and in_info.mask is None:
            with ErrorLogger(self.logger):
                raise FinamMetaDataError("Missing source mask specification")

        if (
            self.output_grid is not None
            and info.grid is not None
            and self.output_grid != info.grid
        ):
            with ErrorLogger(self.logger):
                msg = "Target grid specification is already set, new specs differ"
                raise FinamMetaDataError(msg)

        self.input_grid = self.input_grid or in_info.grid
        self.input_mask = self.input_mask or in_info.mask
        self.output_grid = self.output_grid or info.grid

        if self.input_grid.crs is None and self.output_grid.crs is not None:
            raise FinamMetaDataError("Input grid has a CRS, but output grid does not")
        if self.output_grid.crs is None and self.input_grid.crs is not None:
            raise FinamMetaDataError("output grid has a CRS, but input grid does not")

        if not self._is_initialized:
            self.downstream_mask = info.mask
            self.transformer = _create_transformer(self.output_grid, self.input_grid)
            self._update_grid_specs()
            # self.output_mask may be determined by "_update_grid_specs"
            self._check_and_set_out_mask()
            self._is_initialized = True

        self.input_meta = in_info.meta
        return in_info.copy_with(grid=self.output_grid, mask=self.output_mask)

    def _do_transform(self, points):
        if self.transformer is None:
            return points
        return np.asarray(list(self.transformer.itransform(points)))

    def _check_and_set_out_mask(self):
        if self._out_mask_checked:
            return  # already done
        if (
            self.output_mask is not None
            and self.downstream_mask is not None
            and not dtools.masks_compatible(
                self.output_mask, self.downstream_mask, True
            )
        ):
            with ErrorLogger(self.logger):
                msg = (
                    "Regrid: Target mask specification is already set, new specs differ"
                )
                raise FinamMetaDataError(msg)
        self.output_mask = (
            self.output_mask if self.output_mask is not None else self.downstream_mask
        )
        self._out_mask_checked = self.output_mask is not None

    def _need_mask(self, mask):
        return dtools.mask_specified(mask) and mask is not np.ma.nomask

    def _get_in_coords(self):
        if self._need_mask(self.input_mask):
            return self.input_grid.data_points[
                np.logical_not(self.input_mask.ravel(order=self.input_grid.order))
            ]
        return self.input_grid.data_points

    def _get_out_coords(self):
        if not self._out_mask_checked:
            with ErrorLogger(self.logger):
                msg = (
                    "Regrid: Output coordinates weren't checked for mask compatibility"
                )
                raise FinamMetaDataError(msg)
        if self._need_mask(self.output_mask):
            out_data_points = self.output_grid.data_points[
                np.logical_not(self.output_mask.ravel(order=self.output_grid.order))
            ]
        else:
            out_data_points = self.output_grid.data_points
        return self._do_transform(out_data_points)

    def _check_in_data(self, in_data):
        if dtools.is_masked_array(in_data) and not dtools.mask_specified(
            self.input_mask
        ):
            with ErrorLogger(self.logger):
                msg = "For regridding masked input data, you need to explicitly set the mask in the input info."
                raise FinamDataError(msg)


class RegridNearest(ARegridding):
    """Regrid data between two grid specifications with nearest neighbour interpolation.

    See package `finam-regrid <https://finam.pages.ufz.de/finam-regrid/>`_ for more advanced regridding
    using `ESMPy <https://earthsystemmodeling.org/esmpy/>`_.

    .. warning::
        Does currently not support masked input data. Raises a ``NotImplementedError`` in that case.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.RegridNearest()

        adapter = fm.adapters.RegridNearest(
            in_grid=fm.UniformGrid(dims=(20, 10)),
            out_grid=fm.UniformGrid(dims=(10, 5), spacing=(2.0, 2.0, 2.0)),
        )

    Parameters
    ----------
    in_grid : Grid or None (optional)
        Input grid specification. Will be taken from source component if not specified.
    out_grid : Grid or None (optional)
        Output grid specification. Will be taken from target component if not specified.
    out_mask : :any:`Mask` value or valid boolean mask for :any:`MaskedArray` or None, optional
        masking specification of the regridding output. Options:
            * :any:`Mask.FLEX`: data will be unmasked
            * :any:`Mask.NONE`: data will be unmasked and given as plain numpy array
            * valid boolean mask for MaskedArray
            * None: will be determined by connected target
    tree_options : dict
        kwargs for :class:`scipy.spatial.KDTree`
    """

    def __init__(self, in_grid=None, out_grid=None, out_mask=None, tree_options=None):
        super().__init__(in_grid, out_grid, out_mask)
        self.tree_options = tree_options
        self.ids = None

    def _update_grid_specs(self):
        if self.input_grid.dim != self.output_grid.dim:
            msg = "Input grid and output grid have different dimensions"
            raise FinamMetaDataError(msg)
        # out mask not restricted by nearest interpolation
        self._check_and_set_out_mask()
        # generate IDs to select data
        kw = self.tree_options or {}
        tree = KDTree(self._get_in_coords(), **kw)
        # only store IDs, since they will be constant
        self.ids = tree.query(self._get_out_coords())[1]

    def _get_data(self, time, target):
        in_data = self.pull_data(time, target)
        self._check_in_data(in_data)
        return dtools.from_compressed(
            dtools.to_compressed(in_data, order=self.input_grid.order)[self.ids],
            shape=self.output_grid.data_shape,
            order=self.output_grid.order,
            mask=self.output_mask,
        )


class RegridLinear(ARegridding):
    """
    Regrid data between two grid specifications with linear interpolation.

    Uses :class:`scipy.interpolate.RegularGridInterpolator` for structured grids.
    For unstructured grids, :class:`scipy.interpolate.LinearNDInterpolator` is used,
    which performs triangulation internally.
    So the actual topology of the grid is not taken into account.

    See package `finam-regrid <https://finam.pages.ufz.de/finam-regrid/>`_ for more advanced regridding
    using `ESMPy <https://earthsystemmodeling.org/esmpy/>`_.

    .. warning::
        Does currently not support masked input data. Raises a ``NotImplementedError`` in that case.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.RegridLinear()

        adapter = fm.adapters.RegridLinear(
            in_grid=fm.UniformGrid(dims=(20, 10)),
            out_grid=fm.UniformGrid(dims=(10, 5), spacing=(2.0, 2.0, 2.0)),
        )

    Parameters
    ----------
    in_grid : Grid or None (optional)
        Input grid specification. Will be taken from source component if not specified.
    out_grid : Grid or None (optional)
        Output grid specification. Will be taken from target component if not specified.
    out_mask : :any:`Mask` value or valid boolean mask for :any:`MaskedArray` or None, optional
        masking specification of the regridding output. Options:
            * :any:`Mask.FLEX`: data will be unmasked
            * :any:`Mask.NONE`: data will be unmasked and given as plain numpy array
            * valid boolean mask for MaskedArray
            * None: will be determined by connected target
    fill_with_nearest : bool
        Whether out of bounds points should be filled with the nearest value. Default ``False``.
    tree_options : dict
        kwargs for :class:`scipy.spatial.KDTree`
    """

    def __init__(
        self,
        in_grid=None,
        out_grid=None,
        out_mask=None,
        fill_with_nearest=False,
        tree_options=None,
    ):
        super().__init__(in_grid, out_grid, out_mask)
        self.tree_options = tree_options
        self.fill_with_nearest = bool(fill_with_nearest)
        self.ids = None
        self.inter = None
        self.out_ids = None
        self.fill_ids = None
        self.out_coords = None
        self.structured = False

    def _update_grid_specs(self):
        if isinstance(self.input_grid, StructuredGrid) and not self._need_mask(
            self.input_mask
        ):
            self.structured = True
            self.inter = RegularGridInterpolator(
                points=self.input_grid.data_axes,
                values=np.zeros(self.input_grid.data_shape, dtype=np.double),
                bounds_error=False,
            )
        else:
            in_coords = self._get_in_coords()
            self.inter = LinearNDInterpolator(
                points=in_coords,
                values=np.zeros(len(in_coords), dtype=np.double),
            )
        if self.fill_with_nearest:
            # out mask not restricted when filled with nearest
            self._check_and_set_out_mask()
            self.out_coords = self._get_out_coords()
            # check for outliers once
            res = self.inter(self.out_coords)
            self.out_ids = np.isnan(res)
            out_points = self.out_coords[self.out_ids]
            kw = self.tree_options or {}
            tree = KDTree(self._get_in_coords(), **kw)
            self.fill_ids = tree.query(out_points)[1]
        else:
            mask_save = self.output_mask
            # temporarily unmask
            self._out_mask_checked = True
            self.output_mask = np.ma.nomask
            # check for outliers once
            res = self.inter(self._get_out_coords())
            # create mask from outliers
            outlier_mask = np.ma.make_mask(
                dtools.from_compressed(
                    np.isnan(res), self.output_grid.data_shape, self.output_grid.order
                )
            )
            # determine mask from outliers
            if mask_save is None or mask_save is dtools.Mask.FLEX:
                self.output_mask = outlier_mask
            elif mask_save is dtools.Mask.NONE:
                if np.any(outlier_mask):
                    msg = "RegridLinear: interpolation is not covering desired domain."
                    raise FinamDataError(msg)
                self.output_mask = mask_save
            else:
                if not dtools.is_sub_mask(outlier_mask, mask_save):
                    msg = "RegridLinear: interpolation is not covering desired masked domain."
                    raise FinamDataError(msg)
                self.output_mask = mask_save
            self._out_mask_checked = False
            self._check_and_set_out_mask()
            self.out_coords = self._get_out_coords()

    def _get_data(self, time, target):
        in_data = self.pull_data(time, target)
        self._check_in_data(in_data)

        if self.structured:
            self.inter.values = in_data[0, ...].magnitude
            res = self.inter(self.out_coords)
            if self.fill_with_nearest:
                res[self.out_ids] = self.inter.values.flatten(
                    order=self.input_grid.order
                )[self.fill_ids]
        else:
            in_data = dtools.to_compressed(
                in_data[0, ...].magnitude, order=self.input_grid.order
            )
            self.inter.values = np.ascontiguousarray(
                in_data.reshape((-1, 1)),
                dtype=np.double,
            )
            res = self.inter(self.out_coords)
            if self.fill_with_nearest:
                res[self.out_ids] = self.inter.values[self.fill_ids, 0]
        return dtools.from_compressed(
            res,
            shape=self.output_grid.data_shape,
            order=self.output_grid.order,
            mask=self.output_mask,
        )


def _create_transformer(input_grid, output_grid):
    in_crs = None if input_grid.crs is None else crs.CRS(input_grid.crs)
    out_crs = None if output_grid.crs is None else crs.CRS(output_grid.crs)
    transformer = (
        None
        if (in_crs is None and out_crs is None) or in_crs == out_crs
        else Transformer.from_crs(in_crs, out_crs)
    )
    return transformer



import unittest
from datetime import datetime

import numpy as np

from finam import ComponentStatus, Info, Input, NoGrid, Output, UniformGrid
from finam.sdk.component import IOList
from finam.tools.connect_helper import ConnectHelper, FromInput, FromOutput, FromValue


class TestConnectHelper(unittest.TestCase):
    def test_connect(self):
        time = datetime(2020, 10, 6)
        info = Info(grid=NoGrid(), time=time)

        inputs = IOList(None, "INPUT")
        inputs.add(name="In1")
        inputs.add(name="In2", grid=NoGrid(), time=time)
        outputs = IOList(None, "OUTPUT")
        outputs.add(name="Out1")
        outputs.add(name="Out2")

        sources = [Output("so1"), Output("so1")]
        sinks = [Input("si1"), Input("si2")]

        sources[0] >> inputs["In1"]
        sources[1] >> inputs["In2"]

        outputs["Out1"] >> sinks[0]
        outputs["Out2"] >> sinks[1]

        inputs["In1"].ping()
        inputs["In2"].ping()
        sinks[0].ping()
        sinks[1].ping()

        connector: ConnectHelper = ConnectHelper(
            "TestLogger",
            inputs,
            outputs,
            pull_data=list(inputs.keys()),
            cache=False,
        )
        self.assertEqual(connector.uses_base_logger_name, True)

        self.assertEqual(connector.infos_pushed, {"Out1": False, "Out2": False})
        self.assertEqual(connector.data_pushed, {"Out1": False, "Out2": False})

        self.assertEqual(connector.in_data, {"In1": None, "In2": None})
        self.assertEqual(connector.in_infos, {"In1": None, "In2": None})
        self.assertEqual(connector.out_infos, {"Out1": None, "Out2": None})

        self.assertEqual(connector.data_required, {"Out1": True, "Out2": True})
        self.assertEqual(connector.in_infos_required, {"In1": True, "In2": False})
        self.assertEqual(connector.out_infos_required, {"Out1": True, "Out2": True})

        status = connector.connect(time)

        self.assertEqual(status, ComponentStatus.CONNECTING_IDLE)

        status = connector.connect(time, exchange_infos={"In1": info.copy()})

        sources[0].push_info(info.copy())

        status = connector.connect(time, exchange_infos={"In1": info.copy()})
        self.assertEqual(status, ComponentStatus.CONNECTING)
        self.assertEqual(connector.in_infos, {"In1": info, "In2": None})

        self.assertEqual(connector.data_required, {"Out1": True, "Out2": True})
        self.assertEqual(connector.in_infos_required, {"In1": False, "In2": False})
        self.assertEqual(connector.out_infos_required, {"Out1": True, "Out2": True})

        sources[0].push_data(1, time)

        status = connector.connect(time)
        self.assertEqual(status, ComponentStatus.CONNECTING)
        self.assertEqual(connector.in_data, {"In1": 1, "In2": None})

        status = connector.connect(time)
        self.assertEqual(status, ComponentStatus.CONNECTING_IDLE)

        status = connector.connect(time, push_infos={"Out1": info.copy()})
        self.assertEqual(status, ComponentStatus.CONNECTING)
        self.assertEqual(connector.out_infos, {"Out1": None, "Out2": None})

        self.assertEqual(connector.data_required, {"Out1": True, "Out2": True})
        self.assertEqual(connector.in_infos_required, {"In1": False, "In2": False})
        self.assertEqual(connector.out_infos_required, {"Out1": False, "Out2": True})

        sinks[0].exchange_info(info.copy())

        status = connector.connect(time)
        self.assertEqual(status, ComponentStatus.CONNECTING)
        self.assertEqual(connector.out_infos, {"Out1": info, "Out2": None})

        sources[1].push_info(info.copy())

        status = connector.connect(
            time, exchange_infos={"In2": info.copy()}, push_infos={"Out2": info.copy()}
        )
        sources[1].push_data(2, time)

        self.assertEqual(connector.data_required, {"Out1": True, "Out2": True})
        self.assertEqual(connector.in_infos_required, {"In1": False, "In2": False})
        self.assertEqual(connector.out_infos_required, {"Out1": False, "Out2": False})

        self.assertEqual(status, ComponentStatus.CONNECTING)
        self.assertEqual(connector.in_infos, {"In1": info, "In2": info})

        connector.connect(
            time, exchange_infos={"In2": info.copy()}, push_infos={"Out2": info.copy()}
        )

        self.assertEqual(connector.in_data["In1"], 1)
        self.assertEqual(connector.in_data["In2"], 2)

        sinks[1].exchange_info(info.copy())
        status = connector.connect(time, push_data={"Out1": 1, "Out2": 2})
        self.assertEqual(status, ComponentStatus.CONNECTED)

        self.assertEqual(connector.data_required, {"Out1": False, "Out2": False})

        self.assertEqual(connector.infos_pushed, {"Out1": True, "Out2": True})
        self.assertEqual(connector.data_pushed, {"Out1": True, "Out2": True})

        self.assertEqual(connector.in_data, {"In1": 1, "In2": 2})
        self.assertEqual(connector.in_infos, {"In1": info, "In2": info})
        self.assertEqual(connector.out_infos, {"Out1": info, "Out2": info})

    def test_connect_fail(self):
        time = datetime(2020, 10, 6)

        inputs = IOList(None, "INPUT")
        inputs.add(name="In1")
        inputs.add(name="In2")
        outputs = IOList(None, "OUTPUT")
        outputs.add(name="Out1")
        outputs.add(name="Out2")

        with self.assertRaises(KeyError):
            _connector = ConnectHelper("TestLogger", inputs, outputs, pull_data=["In3"])

        connector = ConnectHelper("TestLogger", inputs, outputs)

        with self.assertRaises(KeyError):
            connector.connect(
                start_time=None, exchange_infos={"In3": Info(time, NoGrid())}
            )
        with self.assertRaises(KeyError):
            connector.connect(
                start_time=None, push_infos={"Out3": Info(time, NoGrid())}
            )
        with self.assertRaises(KeyError):
            connector.connect(start_time=None, push_data={"Out3": 0.0})

    def test_connect_caching(self):
        time = datetime(2020, 10, 6)

        info = Info(grid=NoGrid(), time=time)

        inputs = IOList(None, "INPUT")
        inputs.add(name="In1")
        inputs.add(name="In2")
        outputs = IOList(None, "OUTPUT")
        outputs.add(name="Out1")
        outputs.add(name="Out2")

        sources = [Output("so1"), Output("so1")]
        sinks = [Input("si1"), Input("si2")]

        sources[0] >> inputs["In1"]
        sources[1] >> inputs["In2"]

        outputs["Out1"] >> sinks[0]
        outputs["Out2"] >> sinks[1]

        inputs["In1"].ping()
        inputs["In2"].ping()
        sinks[0].ping()
        sinks[1].ping()

        connector: ConnectHelper = ConnectHelper(
            "TestLogger",
            inputs,
            outputs,
            pull_data=list(inputs.keys()),
            cache=True,
        )
        connector.connect(
            time,
            exchange_infos={"In1": info.copy(), "In2": info.copy()},
            push_infos={"Out1": info.copy(), "Out2": info.copy()},
            push_data={"Out1": 1, "Out2": 2},
        )

        sources[0].push_info(info.copy())
        sources[1].push_info(info.copy())

        sinks[0].exchange_info(info.copy())
        sinks[1].exchange_info(info.copy())

        connector.connect(time)

        sources[0].push_data(1, time)
        sources[1].push_data(2, time)

        status = connector.connect(time)

        self.assertEqual(status, ComponentStatus.CONNECTED)

    def test_connect_transfer_from_input(self):
        time = datetime(2020, 10, 6)

        inputs = IOList(None, "INPUT")
        inputs.add(name="In1", time=None, grid=None)
        inputs.add(name="In2", time=None, grid=None, units=None)
        outputs = IOList(None, "OUTPUT")
        outputs.add(name="Out1")

        sources = [Output("so1"), Output("so1")]
        sink = Input("si1")

        sources[0] >> inputs["In1"]
        sources[1] >> inputs["In2"]

        outputs["Out1"] >> sink

        inputs["In1"].ping()
        inputs["In2"].ping()
        sink.ping()

        connector: ConnectHelper = ConnectHelper(
            "TestLogger",
            inputs,
            outputs,
            pull_data=list(inputs.keys()),
            out_info_rules={
                "Out1": [
                    FromInput("In1", ["time", "grid"]),
                    FromInput("In2", ["units"]),
                    FromValue("test_prop", 1),
                    FromValue("time", time),
                ]
            },
            cache=True,
        )
        connector.connect(
            start_time=None,
            push_data={"Out1": np.zeros((9, 9))},
        )

        sources[0].push_info(Info(time=time, grid=UniformGrid((10, 10))))
        sources[1].push_info(Info(time=time, grid=NoGrid(), units="m"))

        connector.connect(start_time=time)

        self.assertEqual(
            connector.in_infos,
            {
                "In1": Info(time=time, grid=UniformGrid((10, 10))),
                "In2": Info(time=time, grid=NoGrid(), units="m"),
            },
        )
        self.assertEqual(connector.out_infos, {"Out1": None})
        self.assertEqual(connector.infos_pushed, {"Out1": False})

        connector.connect(start_time=time)
        self.assertEqual(connector.infos_pushed, {"Out1": True})
        self.assertEqual(connector.out_infos, {"Out1": None})

        sink.exchange_info(Info(time=time, grid=None, units=None))
        connector.connect(start_time=time)

        self.assertEqual(
            connector.out_infos,
            {
                "Out1": Info(
                    time=time, grid=UniformGrid((10, 10)), units="m", test_prop=1
                )
            },
        )

    def test_connect_transfer_from_output(self):
        time = datetime(2020, 10, 6)

        inputs = IOList(None, "INPUT")
        inputs.add(name="In1")
        inputs.add(name="In2")
        outputs = IOList(None, "OUTPUT")
        outputs.add(name="Out1", time=None, grid=None, units=None)

        sources = [Output("so1"), Output("so1")]
        sink = Input("si1")

        sources[0] >> inputs["In1"]
        sources[1] >> inputs["In2"]

        outputs["Out1"] >> sink

        inputs["In1"].ping()
        inputs["In2"].ping()
        sink.ping()

        connector: ConnectHelper = ConnectHelper(
            "TestLogger",
            inputs,
            outputs,
            pull_data=inputs.names,
            cache=True,
        )
        connector.add_in_info_rule("In1", FromOutput("Out1", ["time", "units"]))
        connector.add_in_info_rule("In2", FromOutput("Out1", ["time", "units"]))
        connector.add_in_info_rule("In2", FromValue("grid", NoGrid()))
        connector.add_in_info_rule("In2", FromValue("time", time))
        connector.connect(
            start_time=time,
            push_data={"Out1": np.zeros((9, 9))},
        )

        sink.exchange_info(Info(time=time, grid=UniformGrid((10, 10)), units="m"))
        connector.connect(start_time=time)

        sources[0].push_info(Info(time=time, grid=UniformGrid((10, 10)), units="m"))
        sources[1].push_info(Info(time=time, grid=NoGrid(), units="m"))

        connector.connect(start_time=time)

        self.assertEqual(
            connector.in_infos,
            {
                "In1": Info(time=time, grid=UniformGrid((10, 10)), units="m"),
                "In2": Info(time=time, grid=NoGrid(), units="m"),
            },
        )

        self.assertEqual(
            connector.out_infos,
            {"Out1": Info(time=time, grid=UniformGrid((10, 10)), units="m")},
        )

    def test_connect_constructor_fail(self):
        inputs = IOList(None, "INPUT")
        inputs.add(name="In1")
        outputs = IOList(None, "OUTPUT")
        outputs.add(name="Out1")

        with self.assertRaises(KeyError):
            _connector: ConnectHelper = ConnectHelper(
                "TestLogger",
                inputs,
                outputs,
                in_info_rules={"InX": []},
            )

        with self.assertRaises(KeyError):
            _connector: ConnectHelper = ConnectHelper(
                "TestLogger",
                inputs,
                outputs,
                out_info_rules={"OutX": []},
            )

        with self.assertRaises(KeyError):
            _connector: ConnectHelper = ConnectHelper(
                "TestLogger",
                inputs,
                outputs,
                in_info_rules={"In1": [FromInput("InX", ["grid"])]},
            )

        with self.assertRaises(KeyError):
            _connector: ConnectHelper = ConnectHelper(
                "TestLogger",
                inputs,
                outputs,
                in_info_rules={"In1": [FromOutput("OutX", ["grid"])]},
            )

        with self.assertRaises(TypeError):
            _connector: ConnectHelper = ConnectHelper(
                "TestLogger",
                inputs,
                outputs,
                in_info_rules={"In1": [FromOutput("OutX", "grid")]},
            )

        with self.assertRaises(TypeError):
            _connector: ConnectHelper = ConnectHelper(
                "TestLogger",
                inputs,
                outputs,
                in_info_rules={"In1": [0]},
            )

    def test_connect_rules_fail(self):
        inputs = IOList(None, "INPUT")
        inputs.add(name="In1")
        outputs = IOList(None, "OUTPUT")
        outputs.add(name="Out1")

        connector: ConnectHelper = ConnectHelper(
            "TestLogger",
            inputs,
            outputs,
        )
        connector.add_in_info_rule("In1", FromOutput("Out1"))
        connector.add_out_info_rule("Out1", FromInput("In1"))
        with self.assertRaises(ValueError):
            connector.connect(
                start_time=None, exchange_infos={"In1": Info(time=None, grid=NoGrid())}
            )

        with self.assertRaises(ValueError):
            connector.connect(
                start_time=None, push_infos={"Out1": Info(time=None, grid=NoGrid())}
            )


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from pathlib import Path

from finam.tools.cwd_helper import execute_in_cwd


class Mockup:
    def __init__(self, cwd=None):
        self.cwd = cwd
        self.new_cwd = None

    @execute_in_cwd
    def change_cwd(self):
        self.new_cwd = os.getcwd()


class TestCWD(unittest.TestCase):
    def test_cwd(self):
        cwd = os.getcwd()
        # no cwd set
        test1 = Mockup()
        with self.assertRaises(ValueError):
            test1.change_cwd()
        # assert we were in the right cwd
        test2 = Mockup("..")
        test2.change_cwd()
        self.assertEqual(Path(test2.cwd).resolve(), Path(test2.new_cwd).resolve())
        self.assertEqual(Path(".").resolve(), Path(cwd).resolve())
        self.assertNotEqual(Path(".").resolve(), Path(test2.cwd).resolve())


if __name__ == "__main__":
    unittest.main()



"""
Specialized data types for exchanges between models/components.

Grids
=====

.. autosummary::
   :toctree: generated

    :noindex: EsriGrid
    :noindex: NoGrid
    :noindex: RectilinearGrid
    :noindex: UniformGrid
    :noindex: UnstructuredGrid
    :noindex: UnstructuredPoints


Grid tools
==========

.. autosummary::
   :toctree: generated

    check_axes_monotonicity
    check_axes_uniformity
    check_uniformity
    :noindex: CellType
    :noindex: Location


Grid abstract base classes
==========================

.. autosummary::
   :toctree: generated

    GridBase
    Grid
    StructuredGrid


Data tools
==========

.. autosummary::
   :toctree: generated

    :noindex: Info
    prepare
    strip_time
    check
    full
    full_like
    has_time_axis
    to_datetime
    assert_type


Unit tools
==========

.. autosummary::
   :toctree: generated

    :noindex: UNITS
    quantify
    to_units
    is_quantified
    check_quantified
    get_dimensionality
    get_magnitude
    get_units


Mask tools
==========

.. autosummary::
   :toctree: generated

    :noindex: Mask
    is_masked_array
    has_masked_values
    filled
    to_masked
    to_compressed
    from_compressed
    check_data_covers_domain
    masks_compatible
    masks_equal
    mask_specified
"""

from ..errors import FinamDataError
from . import grid_base, grid_spec, grid_tools, tools
from .grid_base import Grid, GridBase, StructuredGrid
from .grid_spec import (
    EsriGrid,
    NoGrid,
    RectilinearGrid,
    UniformGrid,
    UnstructuredGrid,
    UnstructuredPoints,
)
from .grid_tools import (
    CellType,
    Location,
    check_axes_monotonicity,
    check_axes_uniformity,
    check_uniformity,
)
from .tools import (
    UNITS,
    Info,
    Mask,
    assert_type,
    check,
    check_data_covers_domain,
    check_quantified,
    filled,
    from_compressed,
    full,
    full_like,
    get_dimensionality,
    get_magnitude,
    get_units,
    has_masked_values,
    has_time_axis,
    is_masked_array,
    is_quantified,
    mask_specified,
    masks_compatible,
    masks_equal,
    prepare,
    quantify,
    strip_time,
    to_compressed,
    to_datetime,
    to_masked,
    to_units,
)

__all__ = ["grid_base", "grid_spec", "grid_tools", "tools"]
__all__ += [
    "EsriGrid",
    "NoGrid",
    "RectilinearGrid",
    "UniformGrid",
    "UnstructuredGrid",
    "UnstructuredPoints",
]
__all__ += [
    "CellType",
    "Grid",
    "GridBase",
    "Location",
    "StructuredGrid",
    "check_axes_monotonicity",
    "check_axes_uniformity",
    "check_uniformity",
]
__all__ += [
    "UNITS",
    "FinamDataError",
    "Info",
    "Mask",
    "assert_type",
    "check",
    "check_quantified",
    "full",
    "full_like",
    "get_dimensionality",
    "get_magnitude",
    "get_units",
    "has_time_axis",
    "prepare",
    "quantify",
    "is_quantified",
    "strip_time",
    "to_datetime",
    "to_units",
]
__all__ += [
    "is_masked_array",
    "has_masked_values",
    "filled",
    "to_masked",
    "to_compressed",
    "from_compressed",
    "check_data_covers_domain",
    "masks_compatible",
    "masks_equal",
    "mask_specified",
]



"""
Unit tests for data info propagation.
"""

import unittest
from datetime import datetime, timedelta

import numpy as np

from finam import (
    UNITS,
    CellType,
    Composition,
    FinamDataError,
    FinamMetaDataError,
    Info,
    Location,
    RectilinearGrid,
    UniformGrid,
    UnstructuredGrid,
)
from finam import data as fdata
from finam.adapters.regrid import RegridLinear, RegridNearest
from finam.components import debug, generators


class TestRegrid(unittest.TestCase):
    def test_regrid_nearest(self):
        time = datetime(2000, 1, 1)

        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(3.0, 3.0, 3.0), data_location=Location.POINTS
            ),
            units="m",
        )

        out_spec = UniformGrid(dims=(14, 29), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: in_data,
                    in_info,
                )
            },
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = Composition([source, sink], log_level="DEBUG")

        (source.outputs["Output"] >> RegridNearest() >> sink.inputs["Input"])

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.grid, out_spec)
        self.assertEqual(sink.data["Input"][0, 0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 0, 1], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 1], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 0, 2], 0.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 2, 0], 0.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 2, 2], 0.0 * UNITS.meter)

    def test_regrid_nearest_masked(self):
        time = datetime(2000, 1, 1)

        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(3.0, 3.0, 3.0), data_location=Location.POINTS
            ),
            units="m",
        )

        out_spec = UniformGrid(dims=(14, 29), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        in_data = np.ma.masked_where(in_data > 0, in_data)

        source = generators.CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: in_data.copy(),
                    in_info,
                )
            },
            start=time,
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=time,
            step=timedelta(days=1),
        )

        composition = Composition([source, sink], log_level="DEBUG")

        (source.outputs["Output"] >> RegridNearest() >> sink.inputs["Input"])

        with self.assertRaises(FinamDataError):
            composition.connect()

    def test_regrid_nearest_crs(self):
        time = datetime(2000, 1, 1)
        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10),
                spacing=(3.0, 3.0, 3.0),
                data_location=Location.POINTS,
                crs="EPSG:32632",
            ),
            units="m",
        )

        out_spec = UniformGrid(
            dims=(14, 29), data_location=Location.POINTS, crs="EPSG:25832"
        )

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: in_data,
                    in_info,
                )
            },
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = Composition([source, sink])

        (source.outputs["Output"] >> RegridNearest() >> sink.inputs["Input"])

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.grid, out_spec)
        self.assertEqual(sink.data["Input"][0, 0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 0, 1], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 1], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 0, 2], 0.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 2, 0], 0.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 2, 2], 0.0 * UNITS.meter)

    def test_regrid_linear(self):
        time = datetime(2000, 1, 1)
        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10),
                spacing=(2.0, 2.0, 2.0),
                data_location=Location.POINTS,
            ),
            units="m",
        )
        out_spec = UniformGrid(dims=(9, 19), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = Composition([source, sink])

        (source.outputs["Output"] >> RegridLinear() >> sink.inputs["Input"])

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.grid, out_spec)
        self.assertEqual(sink.data["Input"][0, 0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 0, 1], 0.5 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 0], 0.5 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 1], 0.25 * UNITS.meter)

    def test_regrid_linear_masked(self):
        time = datetime(2000, 1, 1)

        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(3.0, 3.0, 3.0), data_location=Location.POINTS
            ),
            units="m",
        )

        out_spec = UniformGrid(dims=(14, 29), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        in_data = np.ma.masked_where(in_data > 0, in_data)

        source = generators.CallbackGenerator(
            callbacks={
                "Output": (
                    lambda t: in_data.copy(),
                    in_info,
                )
            },
            start=time,
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=time,
            step=timedelta(days=1),
        )

        composition = Composition([source, sink], log_level="DEBUG")

        (
            source.outputs["Output"]
            >> RegridLinear(fill_with_nearest=True)
            >> sink.inputs["Input"]
        )

        with self.assertRaises(FinamDataError):
            composition.connect()

    def test_regrid_linear_crs(self):
        time = datetime(2000, 1, 1)
        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10),
                spacing=(2.0, 2.0, 2.0),
                data_location=Location.POINTS,
                crs="EPSG:32632",
            ),
            units="m",
        )
        out_spec = UniformGrid(
            dims=(9, 19), data_location=Location.POINTS, crs="EPSG:25832"
        )

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = Composition([source, sink])

        (
            source.outputs["Output"]
            >> RegridLinear(fill_with_nearest=True)
            >> sink.inputs["Input"]
        )

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.grid, out_spec)
        self.assertAlmostEqual(fdata.get_magnitude(sink.data["Input"])[0, 0, 0], 1.0)
        self.assertAlmostEqual(fdata.get_magnitude(sink.data["Input"])[0, 0, 1], 0.5)
        self.assertAlmostEqual(fdata.get_magnitude(sink.data["Input"])[0, 1, 0], 0.5)
        self.assertAlmostEqual(fdata.get_magnitude(sink.data["Input"])[0, 1, 1], 0.25)

    def test_regrid_linear_custom(self):
        time = datetime(2000, 1, 1)
        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(2.0, 2.0, 2.0), data_location=Location.POINTS
            ),
            units="m",
        )
        out_spec = UniformGrid(dims=(9, 19), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={
                "Output": (lambda t: in_data, Info(time=None, grid=None, units="m"))
            },
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=None, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = Composition([source, sink])

        (
            source.outputs["Output"]
            >> RegridLinear(in_grid=in_info.grid, out_grid=out_spec)
            >> sink.inputs["Input"]
        )

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.grid, out_spec)
        self.assertEqual(sink.data["Input"][0, 0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 0, 1], 0.5 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 0], 0.5 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 1], 0.25 * UNITS.meter)

    def test_regrid_linear_rev(self):
        time = datetime(2000, 1, 1)
        in_info = Info(
            time=time,
            grid=RectilinearGrid(
                axes=[np.linspace(8, 0, 5), np.linspace(0, 18, 10)],
                data_location=Location.POINTS,
            ),
            units="m",
        )
        out_spec = UniformGrid(dims=(9, 19), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = Composition([source, sink])

        (source.outputs["Output"] >> RegridLinear() >> sink.inputs["Input"])

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.grid, out_spec)
        self.assertEqual(sink.data["Input"][0, 0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 0, 1], 0.5 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 0], 0.5 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1, 1], 0.25 * UNITS.meter)

    def test_regrid_multi(self):
        time = datetime(2000, 1, 1)
        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(2.0, 2.0, 2.0), data_location=Location.POINTS
            ),
            units="m",
        )
        out_spec_1 = UniformGrid(dims=(9, 19), data_location=Location.POINTS)
        out_spec_2 = UniformGrid(dims=(9, 19), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink_1 = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec_1, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )
        sink_2 = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec_2, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        regrid = RegridLinear()

        composition = Composition([source, sink_1, sink_2])

        source.outputs["Output"] >> regrid
        regrid >> sink_1.inputs["Input"]
        regrid >> sink_2.inputs["Input"]

        composition.connect()

        self.assertEqual(sink_1.inputs["Input"].info.grid, out_spec_1)
        self.assertEqual(sink_1.data["Input"][0, 0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink_1.data["Input"][0, 0, 1], 0.5 * UNITS.meter)
        self.assertEqual(sink_1.data["Input"][0, 1, 0], 0.5 * UNITS.meter)
        self.assertEqual(sink_1.data["Input"][0, 1, 1], 0.25 * UNITS.meter)

        self.assertEqual(sink_2.inputs["Input"].info.grid, out_spec_2)
        self.assertEqual(sink_2.data["Input"][0, 0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink_2.data["Input"][0, 0, 1], 0.5 * UNITS.meter)
        self.assertEqual(sink_2.data["Input"][0, 1, 0], 0.5 * UNITS.meter)
        self.assertEqual(sink_2.data["Input"][0, 1, 1], 0.25 * UNITS.meter)

    def test_regrid_multi_fail(self):
        time = datetime(2000, 1, 1)
        in_info = Info(
            time=time,
            grid=UniformGrid(
                dims=(5, 10), spacing=(2.0, 2.0, 2.0), data_location=Location.POINTS
            ),
            units="m",
        )
        out_spec_1 = UniformGrid(dims=(9, 19), data_location=Location.POINTS)
        out_spec_2 = UniformGrid(dims=(8, 18), data_location=Location.POINTS)

        in_data = np.zeros(shape=in_info.grid.data_shape, order=in_info.grid.order)
        in_data.data[0, 0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink_1 = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec_1, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )
        sink_2 = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec_2, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        regrid = RegridLinear()

        composition = Composition([source, sink_1, sink_2])

        source.outputs["Output"] >> regrid
        regrid >> sink_1.inputs["Input"]
        regrid >> sink_2.inputs["Input"]

        with self.assertRaises(FinamMetaDataError) as _context:
            composition.run(end_time=datetime(2000, 1, 2))

    def test_regrid_linear_unstructured(self):
        time = datetime(2000, 1, 1)
        g1 = UniformGrid(
            dims=(5, 10),
            spacing=(2.0, 2.0, 2.0),
            data_location=Location.POINTS,
        )
        g2 = UniformGrid(dims=(9, 19), data_location=Location.POINTS)

        in_info = Info(
            time=time,
            grid=UnstructuredGrid(
                points=g1.data_points,
                cells=g1.cells,
                cell_types=[CellType.QUAD] * g1.data_size,
                data_location="POINTS",
            ),
            units="m",
        )
        out_spec = UnstructuredGrid(
            points=g2.data_points,
            cells=g2.cells,
            cell_types=[CellType.QUAD] * g2.data_size,
            data_location="POINTS",
        )

        in_data = np.zeros(shape=in_info.grid.data_size)
        in_data.data[0] = 1.0

        source = generators.CallbackGenerator(
            callbacks={"Output": (lambda t: in_data, in_info)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        sink = debug.DebugConsumer(
            {"Input": Info(None, grid=out_spec, units=None)},
            start=datetime(2000, 1, 1),
            step=timedelta(days=1),
        )

        composition = Composition([source, sink])

        (
            source.outputs["Output"]
            >> RegridLinear(fill_with_nearest=True)
            >> sink.inputs["Input"]
        )

        composition.connect()

        self.assertEqual(sink.inputs["Input"].info.grid, out_spec)
        self.assertEqual(sink.data["Input"][0, 0], 1.0 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 1], 0.5 * UNITS.meter)
        self.assertEqual(sink.data["Input"][0, 9], 0.5 * UNITS.meter)


if __name__ == "__main__":
    unittest.main()



"""Adapters for time integration"""

from abc import ABC
from datetime import timedelta

from ..data import tools
from ..errors import FinamNoDataError, FinamTimeError
from ..tools.log_helper import ErrorLogger
from .time import TimeCachingAdapter, check_time, interpolate


class TimeIntegrationAdapter(TimeCachingAdapter, ABC):
    """Abstract base class for time integration adapters."""

    def __init__(self):
        super().__init__()
        self._prev_time = None

    def _source_updated(self, time):
        """Informs the input that a new output is available.

        Parameters
        ----------
        time : datetime
            Simulation time of the notification.
        """
        check_time(self.logger, time)

        data = tools.strip_time(self.pull_data(time, self), self._input_info.grid)
        self.data.append((time, self._pack(data)))

        if self._prev_time is None:
            self._prev_time = time

    def _get_data(self, time, _target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        if len(self.data) == 0:
            raise FinamNoDataError(f"No data available in {self.name}")

        check_time(self.logger, time, (self.data[0][0], self.data[-1][0]))

        sum_value = self._interpolate(time)
        self._clear_cached_data(self._prev_time)
        self._prev_time = time
        return sum_value


# pylint: disable=too-many-ancestors
class AvgOverTime(TimeIntegrationAdapter):
    """Aggregates data over time to form the temporal average over the last pull time step.

    Output is the average height of the Area under Curve (AuC)
    between the last and the current pull (vertical lines):

    .. plot:: api/plots/integration-methods.py

        Illustration of time integration.

    Can use step-wise or linear interpolation between push time steps.

    .. plot:: api/plots/interpolation-methods.py

        Illustration of interpolation methods.

    Parameters
    ----------

    step : float, optional
        Value in range [0, 1] that determines the relative step position.

        Linear interpolation is used if set to ``None`` (the default).

        * For a value of 0.0, the new value is returned for any dt > 0.0.
        * For a value of 1.0, the old value is returned for any dt <= 1.0.
        * Values between 0.0 and 1.0 shift the step between the first and the second time.
        * A value of 0.5 results in nearest interpolation.

    See also
    --------

    .adapters.LinearTime : Linear time interpolation.
    .adapters.StepTime : Step-wise time interpolation.
    .adapters.SumOverTime : Sum aggregation over time.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.AvgOverTime()
    """

    def __init__(self, step=None):
        super().__init__()
        self._prev_time = None
        self._step = step

    def _interpolate(self, time):
        if len(self.data) == 1:
            return self._unpack(self.data[0][1])

        if time <= self.data[0][0]:
            return self._unpack(self.data[0][1])

        sum_value = None

        t_old, v_old = self.data[0]
        v_old = self._unpack(v_old)
        for i in range(len(self.data) - 1):
            t_new, v_new = self.data[i + 1]
            v_new = self._unpack(v_new)

            if self._prev_time >= t_new:
                t_old, v_old = t_new, v_new
                continue
            if time <= t_old:
                break

            time_range = t_new - t_old

            dt1 = max((self._prev_time - t_old) / time_range, 0.0)
            dt2 = min((time - t_old) / time_range, 1.0)

            if self._step is None:
                v1 = interpolate(v_old, v_new, dt1)
                v2 = interpolate(v_old, v_new, dt2)
                value = (dt2 - dt1) * 0.5 * (v1 + v2)
            else:
                dt1_c = min(dt1, self._step)
                dt2_c = max(self._step, dt2)
                value = (min(self._step, dt2) - dt1_c) * v_old + (
                    dt2_c - max(self._step, dt1)
                ) * v_new

            value *= time_range.total_seconds() * tools.UNITS.Unit("s")

            sum_value = value if sum_value is None else sum_value + value

            t_old, v_old = t_new, v_new

        dt = time - self._prev_time
        if dt.total_seconds() > 0:
            sum_value /= dt.total_seconds() * tools.UNITS.Unit("s")
        else:
            with ErrorLogger(self.logger):
                raise FinamTimeError(
                    "Can't calculate average over zero-length time duration."
                )

        return sum_value


# pylint: disable=too-many-ancestors
class SumOverTime(TimeIntegrationAdapter):
    """Aggregates data over time to form the temporal sum (area under curve) over the last pull time step.

    Output is the Area under Curve (AuC) between the last and the current pull (vertical lines):

    .. plot:: api/plots/integration-methods.py

        Illustration of time integration.

    Can use step-wise or linear interpolation between push time steps.

    .. plot:: api/plots/interpolation-methods.py

        Illustration of interpolation methods.

    Parameters
    ----------

    step : float, optional
        Value in range [0, 1] that determines the relative step position.
        Defaults to 0.0, which means values are interpreted as constant over each time step.

        Linear interpolation is used if set to ``None``.

        * For a value of 0.0, the new value is returned for any dt > 0.0.
        * For a value of 1.0, the old value is returned for any dt <= 1.0.
        * Values between 0.0 and 1.0 shift the step between the first and the second time.
        * A value of 0.5 results in nearest interpolation.
    per_time : bool, optional
        Whether the input data is time-normalized.

        Use ``True`` for input units like mm/d, e.g. for precipitation.
        Data per step is multiplied with step length, with time cancelling out:
        mm/d becomes mm.

        Use ``False`` for absolute amount, e.g. mm for the precipitation of the last time step.
        Output units will be the same as input units. Duration of time steps is not explicitly taken into account.

        Examples:

        * ``per_time=True``, value=1mm/d, step=2x5d --> 10 mm
        * ``per_time=True``, value=1mm, step=2x5d --> 10mm*d
        * ``per_time=False``, value=1mm/d, step=2x5d --> 2mm/d
        * ``per_time=False``, value=1mm, step=2x5d --> 2mm

    initial_interval: :class:`datetime <datetime.datetime>`, optional
        Time scaling duration for the initial data, if ``per_time=True``. Defaults to 0 days.

    See also
    --------

    .adapters.LinearTime : Linear time interpolation.
    .adapters.StepTime : Step-wise time interpolation.
    .adapters.AvgOverTime : Average aggregation over time.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.SumOverTime()

    """

    def __init__(self, step=0.0, per_time=True, initial_interval=timedelta(0)):
        super().__init__()
        self._prev_time = None
        self._step = step
        self._per_time = per_time
        self._initial_interval = initial_interval

    def _interpolate(self, time):
        if len(self.data) == 1 or time <= self.data[0][0]:
            if self._per_time:
                return (
                    self._unpack(self.data[0][1])
                    * (self._initial_interval.total_seconds() * tools.UNITS.Unit("s"))
                ).to_reduced_units()

            return self._unpack(self.data[0][1])

        sum_value = None

        t_old, v_old = self.data[0]
        v_old = self._unpack(v_old)
        for i in range(len(self.data) - 1):
            t_new, v_new = self.data[i + 1]
            v_new = self._unpack(v_new)

            if self._prev_time >= t_new:
                t_old, v_old = t_new, v_new
                continue
            if time <= t_old:
                break

            time_range = t_new - t_old

            dt1 = max((self._prev_time - t_old) / time_range, 0.0)
            dt2 = min((time - t_old) / time_range, 1.0)

            if self._step is None:
                v1 = interpolate(v_old, v_new, dt1)
                v2 = interpolate(v_old, v_new, dt2)
                value = (dt2 - dt1) * 0.5 * (v1 + v2)
            else:
                dt1_c = min(dt1, self._step)
                dt2_c = max(self._step, dt2)
                value = (min(self._step, dt2) - dt1_c) * v_old + (
                    dt2_c - max(self._step, dt1)
                ) * v_new

            if self._per_time:
                value *= time_range.total_seconds() * tools.UNITS.Unit("s")

            sum_value = value if sum_value is None else sum_value + value

            t_old, v_old = t_new, v_new

        if self._per_time:
            return sum_value.to_reduced_units()

        return sum_value

    def _get_info(self, info):
        if self._per_time:
            up_info = info.copy_with(units=None)
        else:
            up_info = info.copy_with()

        in_info = self.exchange_info(up_info)

        units = in_info.units
        if self._per_time:
            units *= tools.UNITS.Unit("s")
            out_info = in_info.copy_with(units=(1.0 * units).to_reduced_units().units)
        else:
            out_info = in_info.copy_with()

        return out_info



"""
Modules for reading data.
"""
# pylint: disable=E1101
from datetime import datetime

from finam.interfaces import ComponentStatus

from ..data.grid_spec import NoGrid
from ..data.tools import Info, quantify
from ..sdk import TimeComponent


class CsvReader(TimeComponent):
    """Reads CSV time series with one row per time step, and emits values based on a time column.

    .. code-block:: text

        +-----------+
        |           | [custom] -->
        | CsvReader | [custom] -->
        |           | [......] -->
        +-----------+

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        reader = fm.components.CsvReader(
            path="test.csv",
            time_column="T",
            outputs={
                "A": "meter",
                "B": "",
            },
            date_format=None,
            separator=",",
        )

    .. testcode:: constructor
        :hide:

        reader.initialize()

    Parameters
    ----------
    path : PathLike
        Path to the input file.
    time_column
        Time column name.
    outputs : dict(str, str)
        Data column names and units. Names become output names.
        Use ``""`` for dimensionless columns and ``None`` to get units from connected components.
    date_format : optional
        Format specifier for date.
        See `datetime.strftime <https://docs.python.org/3/library/time.html#time.strftime>`_ for details.
        Default is ISO format.
    separator : str
        Columns separator. Default ";"
    """

    def __init__(self, path, time_column, outputs, date_format=None, separator=";"):
        super().__init__()
        self._path = path
        self._time = None
        self._time_column = time_column
        self._date_format = date_format
        self._separator = separator
        self._data = None
        self._row_index = 0
        self._data_generated = False

        self._output_units = outputs

    def _next_time(self):
        return None

    def _initialize(self):
        """Initialize the component.

        After the method call, the component's inputs and outputs must be available,
        and the component should have status INITIALIZED.
        """
        for name in self._output_units:
            self.outputs.add(name=name)

        self.create_connector()

    def _connect(self, start_time):
        """Push initial values to outputs.

        After the method call, the component should have status CONNECTED.
        """
        import pandas

        push_data = {}
        if not self._data_generated:
            self._data = pandas.read_csv(self._path, sep=self._separator)
            row = self._data.iloc[0]

            self._time, out_data = self._push_row(row, False)
            self._row_index = 1

            if all(
                info is not None for _name, info in self.connector.out_infos.items()
            ):
                push_data = out_data
                self._data_generated = True

        self.try_connect(
            start_time=start_time,
            push_infos={
                name: Info(self.time, grid=NoGrid(), units=units)
                for name, units in self._output_units.items()
            },
            push_data=push_data,
        )

    def _validate(self):
        """Validate the correctness of the component's settings and coupling.

        After the method call, the component should have status VALIDATED.
        """

    def _update(self):
        """Update the component by one time step.
        Push new values to outputs.

        After the method call, the component should have status UPDATED or FINISHED.
        """
        self._time, _ = self._push_row(self._data.iloc[self._row_index], True)
        self._row_index += 1

        if self._row_index >= self._data.shape[0]:
            self.status = ComponentStatus.FINISHED

    def _finalize(self):
        """Finalize and clean up the component.

        After the method call, the component should have status FINALIZED.
        """

    def _push_row(self, row, push):
        if self._date_format is None:
            time = datetime.fromisoformat(row[self._time_column])
        else:
            time = datetime.strptime(row[self._time_column], self._date_format)

        out_data = {
            name: quantify(row[name], units)
            for name, units in self._output_units.items()
        }

        if push:
            for o in self._output_units:
                self.outputs[o].push_data(out_data[o], time)

        return time, out_data



"""
Implementations of FINAM interfaces for component and adapter development.

Software development kit
========================

.. autosummary::
   :toctree: generated

    :noindex: Adapter
    :noindex: Component
    :noindex: TimeComponent
    :noindex: CallbackInput
    :noindex: CallbackOutput
    :noindex: Input
    :noindex: Output
"""
from .adapter import Adapter, TimeDelayAdapter
from .component import Component, TimeComponent
from .input import CallbackInput, Input
from .output import CallbackOutput, Output

__all__ = [
    "Adapter",
    "Component",
    "TimeComponent",
    "TimeDelayAdapter",
    "CallbackInput",
    "CallbackOutput",
    "Input",
    "Output",
]



"""
Adapters for direct probing from link connections.
"""
from ..sdk import Adapter

__all__ = ["CallbackProbe"]


class CallbackProbe(Adapter):
    """Probe data by calling a callback. Simply forwards the data unchanged.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.CallbackProbe(
            callback=lambda data, t: print(data),
        )

    Parameters
    ----------
    callback : callable
        A callback ``callback(data, time)``, returning the transformed data.
    """

    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def _get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        data = self.pull_data(time, target)
        self.callback(data, time)
        return data



import unittest
from datetime import datetime, timedelta

import numpy as np

import finam as fm
from finam import Composition, Info, NoGrid
from finam.components import DebugPushConsumer, StaticCallbackGenerator


def consume(_inputs, _time):
    return {}


class TestStaticCallbackGenerator(unittest.TestCase):
    def test_static_callback(self):
        source = StaticCallbackGenerator(
            callbacks={
                "Out1": (
                    lambda: 1.0 + np.random.random(1)[0],
                    Info(None, grid=NoGrid()),
                )
            },
        )
        consumer = DebugPushConsumer(
            inputs={"In1": Info(None, grid=NoGrid())},
            log_data="DEBUG",
        )

        composition = Composition([source, consumer])

        _ = source.outputs["Out1"] >> fm.adapters.Scale(2.0) >> consumer.inputs["In1"]

        composition.connect(None)

        out_data_1 = consumer.data["In1"]

        self.assertGreaterEqual(out_data_1, 2.0)
        self.assertLessEqual(out_data_1, 4.0)

        composition.run()

        out_data_2 = consumer.data["In1"]
        self.assertEqual(out_data_1, out_data_2)


if __name__ == "__main__":
    unittest.main()



"""
Driver/scheduler for executing a coupled model composition.

Composition
===========

.. autosummary::
   :toctree: generated

    :noindex: Composition
"""

import logging
import os
import sys
from datetime import datetime
from pathlib import Path
from time import strftime

from ._version import __version__
from .errors import (
    FinamCircularCouplingError,
    FinamConnectError,
    FinamStatusError,
    FinamTimeError,
)
from .interfaces import (
    ComponentStatus,
    IAdapter,
    IComponent,
    IInput,
    IOutput,
    ITimeComponent,
    ITimeDelayAdapter,
    Loggable,
    NoBranchAdapter,
    NoDependencyAdapter,
)
from .tools.log_helper import ErrorLogger, is_loggable


class Composition(Loggable):
    """A composition of linked components.

    Manages initialization, initial connection and update schedule of components.

    See :doc:`/finam-book/usage/coupling_scripts` for usage details.

    Examples
    --------

    .. code-block:: Python

        comp_a = SomeComponent(...)
        comp_b = AnotherComponent(...)

        composition = Composition([comp_a, comp_b])

        comp_b.outputs["Out"] >> SomeAdapter() >> comp_b.inputs["In"]

        composition.run(end_time=...)

    Parameters
    ----------
    components : list of IComponent
        Components in the composition.
    logger_name : str, optional
        Name for the base logger, by default "FINAM"
    print_log : bool, optional
        Whether to print log to stdout, by default True
    log_file : str, None or bool, optional
        Whether to write a log file, by default None
    log_level : int or str, optional
        Logging level, by default logging.INFO
    slot_memory_limit : int, optional
        Memory limit per output and adapter data, in bytes.
        When the limit is exceeded, data is stored to disk under the path of ``slot_memory_location``.
        Default: no limit (``None``).
    slot_memory_location : str, optional
        Location for storing data when exceeding ``slot_memory_limit``.
        Default: "temp".
    """

    def __init__(
        self,
        components,
        logger_name="FINAM",
        print_log=True,
        log_file=None,
        log_level=logging.INFO,
        slot_memory_limit=None,
        slot_memory_location="temp",
    ):
        super().__init__()
        # setup logger
        self._logger_name = logger_name
        self.logger.setLevel(log_level)
        # set log format
        formatter = logging.Formatter(
            "%(asctime)s %(levelname)s: %(message)-36s - %(name)s"
        )
        # setup log output
        if print_log:
            sh = logging.StreamHandler(sys.stdout)
            sh.setFormatter(formatter)
            self.logger.addHandler(sh)
        if log_file:
            # for log_file=True use a default name
            if isinstance(log_file, bool):
                log_file = f"./{logger_name}_{strftime('%Y-%m-%d_%H-%M-%S')}.log"
            fh = logging.FileHandler(Path(log_file), mode="w")
            fh.setFormatter(formatter)
            self.logger.addHandler(fh)

        for comp in components:
            if not isinstance(comp, IComponent):
                with ErrorLogger(self.logger):
                    raise ValueError(
                        "Composition: components need to be instances of 'IComponent'."
                    )
        self._components = components
        self._adapters = set()
        self._dependencies = None
        self._input_owners = None
        self._output_owners = None
        self._is_connected = False

        self._time_frame = (None, None)

        self._slot_memory_limit = slot_memory_limit
        self._slot_memory_location = slot_memory_location

        # initialize
        self.logger.info("init composition")

        for comp in self._components:
            self._check_status(comp, [ComponentStatus.CREATED])

        if self._slot_memory_location is not None:
            os.makedirs(self._slot_memory_location, exist_ok=True)

        for comp in self._components:
            if is_loggable(comp) and comp.uses_base_logger_name:
                comp.base_logger_name = self.logger_name
            comp.initialize()
            # set logger
            with ErrorLogger(self.logger):
                comp.inputs.set_logger(comp)
                comp.outputs.set_logger(comp)

            for _, out in comp.outputs.items():
                if out.memory_limit is None:
                    out.memory_limit = self._slot_memory_limit
                if out.memory_location is None:
                    out.memory_location = self._slot_memory_location

            self._check_status(comp, [ComponentStatus.INITIALIZED])

    def connect(self, start_time=None):
        """Performs the connect and validate phases of the composition

        If this was not called by the user, it is called at the start of :meth:`.run`.

        Parameters
        ----------
        start_time : :class:`datetime <datetime.datetime>`, optional
            Starting time of the composition.
            If provided, it should be the starting time of the earliest component.
            If not provided, the composition tries to determine the starting time automatically.
        """
        if self._is_connected:
            raise FinamStatusError("Composition was already connected.")

        time_components = [m for m in self._components if isinstance(m, ITimeComponent)]

        with ErrorLogger(self.logger):
            if len(time_components) == 0:
                if start_time is not None:
                    raise ValueError(
                        "start must be None for a composition without time components"
                    )
            else:
                if start_time is None:
                    start_time = _get_start_time(time_components)
                if not isinstance(start_time, datetime):
                    raise ValueError(
                        "start must be of type datetime for a composition with time components"
                    )

        self._collect_adapters()
        self._validate_composition()

        for ada in self._adapters:
            if ada.memory_limit is None:
                ada.memory_limit = self._slot_memory_limit
            if ada.memory_location is None:
                ada.memory_location = self._slot_memory_location

        self._connect_components(start_time)

        self.logger.info("validate components")
        for comp in self._components:
            comp.validate()
            self._check_status(comp, [ComponentStatus.VALIDATED])

        self._output_owners = _map_outputs(self._components)
        self._input_owners = _map_inputs(self._components)

        self._is_connected = True
        self._time_frame = (start_time, None)

    def run(self, start_time=None, end_time=None):
        """Run this composition using the loop-based update strategy.

        Performs the connect phase if it ``connect()`` was not already called.

        Parameters
        ----------
        start_time : :class:`datetime <datetime.datetime>`, optional
            Starting time of the composition.
            If provided, it should be the starting time of the earliest component.
            If not provided, the composition tries to determine the starting time automatically.
            Ignored if :meth:`.connect` was already called.
        end_time : :class:`datetime <datetime.datetime>`, optional
            Simulation time up to which to simulate.
            Should be ``None`` if no components with time are present.
        """
        time_components = [m for m in self._components if isinstance(m, ITimeComponent)]

        with ErrorLogger(self.logger):
            if len(time_components) == 0:
                if end_time is not None:
                    raise ValueError(
                        "end must be None for a composition without time components"
                    )
            else:
                if not isinstance(end_time, datetime):
                    raise ValueError(
                        "end must be of type datetime for a composition with time components"
                    )

        if not self._is_connected:
            self.connect(start_time)

        self._time_frame = (self._time_frame[0], end_time)

        self.logger.info("run composition")
        while len(time_components) > 0:
            sort_components = list(time_components)
            sort_components.sort(key=lambda m: m.time)
            to_update = sort_components[0]
            updated = self._update_recursive(to_update)
            self._check_status(
                updated, [ComponentStatus.VALIDATED, ComponentStatus.UPDATED]
            )

            any_running = False
            for comp in time_components:
                if comp.status != ComponentStatus.FINISHED and comp.time < end_time:
                    any_running = True
                    break

            if not any_running:
                break

        self._finalize_components()
        self._finalize_composition()

    def _update_recursive(self, comp, chain=None, target_time=None):
        chain = chain or {}
        if comp in chain:
            with ErrorLogger(self.logger):
                joined = " >> ".join(
                    [
                        f"({'*' if delayed else ''}{t or '-'}) {c.name}"
                        for c, (t, delayed) in reversed(chain.items())
                    ]
                )
                raise FinamCircularCouplingError(
                    f"Unresolved circular coupling:\n"
                    f"{comp.name} >> "
                    f"{joined}\n"
                    f"(Deltas are time lags of upstream components, * denotes delayed links)\n"
                    f"You may need to insert a NoDependencyAdapter or ITimeDelayAdapter subclass somewhere, "
                    f"or increase the adapter's delay."
                )

        chain[comp] = None

        if isinstance(comp, ITimeComponent):
            target_time = comp.next_time

        deps = _find_dependencies(comp, self._output_owners, target_time)

        for dep, (local_time, delayed) in deps.items():
            c = self._output_owners[dep]
            if isinstance(c, ITimeComponent):
                if dep.time < local_time:
                    chain[comp] = (local_time - dep.time, delayed)
                    return self._update_recursive(c, chain)
            else:
                updated = self._update_recursive(c, chain, local_time)
                if updated is not None:
                    return updated

        if isinstance(comp, ITimeComponent):
            if comp.status != ComponentStatus.FINISHED:
                comp.update()
            else:
                raise FinamTimeError(
                    f"Can't update dependency component {comp.name}, as it is already finished."
                )
            return comp

        return None

    def _collect_adapters(self):
        for comp in self._components:
            for _, inp in comp.inputs.items():
                _collect_adapters_input(inp, self._adapters)
            for _, out in comp.outputs.items():
                _collect_adapters_output(out, self._adapters)

    def _validate_composition(self):
        """Validates the coupling setup by checking for dangling inputs and disallowed branching connections."""
        self.logger.info("validate composition")
        for comp in self._components:
            with ErrorLogger(comp.logger if is_loggable(comp) else self.logger):
                for inp in comp.inputs.values():
                    _check_input_connected(comp, inp)
                    _check_dead_links(comp, inp)

                for out in comp.outputs.values():
                    _check_branching(comp, out)

        with ErrorLogger(self.logger):
            _check_missing_components(self._components)

    def _connect_components(self, time):
        self.logger.info("connect components")
        counter = 0
        while True:
            self.logger.debug("connect iteration %d", counter)
            any_unconnected = False
            any_new_connection = False
            for comp in self._components:
                if comp.status != ComponentStatus.CONNECTED:
                    comp.connect(time)
                    self._check_status(
                        comp,
                        [
                            ComponentStatus.CONNECTING,
                            ComponentStatus.CONNECTING_IDLE,
                            ComponentStatus.CONNECTED,
                        ],
                    )
                    if comp.status == ComponentStatus.CONNECTED:
                        any_new_connection = True
                    else:
                        if comp.status == ComponentStatus.CONNECTING:
                            any_new_connection = True

                        any_unconnected = True

            if not any_unconnected:
                break
            if not any_new_connection:
                unconn = [
                    m.name
                    for m in self._components
                    if m.status != ComponentStatus.CONNECTED
                ]
                with ErrorLogger(self.logger):
                    raise FinamCircularCouplingError(
                        f"Unresolved circular coupling during initial connect. "
                        f"Unconnected components: [{', '.join(unconn)}]"
                    )

            counter += 1

    def _finalize_components(self):
        self.logger.info("finalize components")
        for comp in self._components:
            self._check_status(
                comp,
                [
                    ComponentStatus.VALIDATED,
                    ComponentStatus.UPDATED,
                    ComponentStatus.FINISHED,
                ],
            )
            if (
                isinstance(comp, ITimeComponent)
                and comp.status == ComponentStatus.VALIDATED
            ):
                self.logger.warning(
                    "Time component %s was not updated during this run", comp.name
                )
            comp.finalize()
            self._check_status(comp, [ComponentStatus.FINALIZED])

        for ada in self._adapters:
            ada.finalize()

    def _finalize_composition(self):
        self.logger.info("finalize composition")
        handlers = self.logger.handlers[:]
        for handler in handlers:
            self.logger.removeHandler(handler)
            handler.close()

    @property
    def logger_name(self):
        """Logger name for the composition."""
        return self._logger_name

    @property
    def uses_base_logger_name(self):
        """Whether this class has a ``base_logger_name`` attribute. False."""
        return False

    def _check_status(self, comp, desired_list):
        if comp.status not in desired_list:
            with ErrorLogger(comp.logger if is_loggable(comp) else self.logger):
                raise FinamStatusError(
                    f"Unexpected component state {comp.status} in {comp.name}. "
                    f"Expecting one of [{', '.join(map(str, desired_list))}]"
                )

    @property
    def metadata(self):
        """
        Meta data for all components and adapters.
        Can only be used after ``connect``.

        Returns
        -------
        dict
            A ``dict`` with the following metadata keys:
              - ``components`` - A `dict` containing metadata for all components.
                Individual entries are generated by :attr:`Component.metadata`
              - ``adapters`` - A `dict` containing metadata for all adapters.
                Individual entries are generated by :attr:`Adapter.metadata`
              - ``links`` - A list of all coupling connections
              - ``time_frame`` - A list of two items: simulation start and end time

            Component and adapter sub-dictionaries use keys like ``name@id``.

        Raises
        ------
        FinamStatusError
            Raises the error if ``connect`` was not called.
        """
        if not self._is_connected:
            with ErrorLogger(self.logger):
                raise FinamStatusError(
                    "can't get meta data for a composition before connect was called"
                )

        comps = {}
        for comp in self._components:
            key = f"{comp.name}@{id(comp)}"
            comps[key] = comp.metadata

        adas = {}
        for ada in self._adapters:
            key = f"{ada.name}@{id(ada)}"
            adas[key] = ada.metadata

        links = []

        for comp in self._components:
            for out_name, out in comp.outputs.items():
                for target in out.targets:
                    if isinstance(target, IAdapter):
                        to = {
                            "adapter": f"{target.name}@{id(target)}",
                        }
                    else:
                        owner = self._input_owners[target]
                        to = {
                            "component": f"{owner.name}@{id(owner)}",
                            "input": target.name,
                        }

                    links.append(
                        {
                            "from": {
                                "component": f"{comp.name}@{id(comp)}",
                                "output": out_name,
                            },
                            "to": to,
                        }
                    )

        for ada in self._adapters:
            for target in ada.targets:
                if isinstance(target, IAdapter):
                    to = {
                        "adapter": f"{target.name}@{id(target)}",
                    }
                else:
                    owner = self._input_owners[target]
                    to = {
                        "component": f"{owner.name}@{id(owner)}",
                        "input": target.name,
                    }

                links.append(
                    {
                        "from": {
                            "adapter": f"{ada.name}@{id(ada)}",
                        },
                        "to": to,
                    }
                )

        return {
            "version": __version__,
            "components": comps,
            "adapters": adas,
            "links": links,
            "time_frame": list(self._time_frame),
        }


def _collect_adapters_input(inp: IInput, out_adapters: set):
    src = inp.source
    if src is None:
        return

    if isinstance(src, IAdapter):
        out_adapters.add(src)
        _collect_adapters_input(src, out_adapters)


def _collect_adapters_output(out: IOutput, out_adapters: set):
    for trg in out.targets:
        if isinstance(trg, IAdapter):
            out_adapters.add(trg)
            _collect_adapters_output(trg, out_adapters)


def _get_start_time(time_components):
    t_min = None
    for comp in time_components:
        if comp.time is not None:
            if t_min is None or comp.time < t_min:
                t_min = comp.time
    if t_min is None:
        raise ValueError(
            "Unable to determine starting time of the composition."
            "Please provide a starting time in ``run()`` or ``connect()``"
        )
    return t_min


def _check_missing_components(components):
    inputs, outputs = _collect_inputs_outputs(components)

    comp_inputs = {inp for comp in components for inp in comp.inputs.values()}
    comp_outputs = {out for comp in components for out in comp.outputs.values()}

    unlinked_inputs = inputs - comp_inputs
    comp_outputs = outputs - comp_outputs

    if len(unlinked_inputs) > 0:
        raise FinamConnectError(
            f"A component was coupled, but not added to this Composition. "
            f"Affected inputs: {[inp.name for inp in unlinked_inputs]}"
        )
    if len(comp_outputs) > 0:
        raise FinamConnectError(
            f"A component was coupled, but not added to this Composition. "
            f"Affected outputs: {[out.name for out in comp_outputs]}"
        )


def _collect_inputs_outputs(components):
    all_inputs = set()
    all_outputs = set()

    for comp in components:
        for _, inp in comp.inputs.items():
            while isinstance(inp, IInput):
                inp = inp.source
            all_outputs.add(inp)

        for _, out in comp.outputs.items():
            targets = {out}
            while len(targets) > 0:
                target = targets.pop()
                curr_targets = target.targets
                for target in curr_targets:
                    if isinstance(target, IOutput):
                        targets.add(target)
                    else:
                        all_inputs.add(target)

    return all_inputs, all_outputs


def _check_branching(comp, out):
    targets = [(out, False)]

    while len(targets) > 0:
        target, no_branch = targets.pop()
        no_branch = no_branch or isinstance(target, NoBranchAdapter)

        curr_targets = target.targets

        if no_branch and len(curr_targets) > 1:
            raise FinamConnectError(
                f"Disallowed branching of output '{out.name}' for "
                f"component {comp.name} ({target.__class__.__name__})"
            )

        for target in curr_targets:
            if isinstance(target, IOutput):
                targets.append((target, no_branch))


def _check_input_connected(comp, inp):
    static = inp.is_static

    while isinstance(inp, IInput):
        if inp.source is None:
            raise FinamConnectError(
                f"Unconnected input '{inp.name}' for target component {comp.name}"
            )
        inp = inp.source

    if static and not inp.is_static:
        raise FinamConnectError("Can't connect a static input to a non-static output.")


def _check_dead_links(comp, inp):
    chain = [inp]
    while isinstance(inp, IInput):
        inp = inp.source
        chain.append(inp)

    first_index = -1
    for i, item in enumerate(reversed(chain)):
        if first_index >= 0 and item.needs_push:
            raise _dead_link_error(comp, chain, first_index, i)
        if item.needs_pull:
            first_index = i


def _map_outputs(components):
    out_map = {}
    for comp in components:
        for _, out in comp.outputs.items():
            out_map[out] = comp
    return out_map


def _map_inputs(components):
    in_map = {}
    for comp in components:
        for _, inp in comp.inputs.items():
            in_map[inp] = comp
    return in_map


def _find_dependencies(component, output_owners, target_time):
    deps = {}
    for _, inp in component.inputs.items():
        local_time = target_time
        delayed = False
        while isinstance(inp, IInput):
            inp = inp.source
            if isinstance(inp, NoDependencyAdapter):
                break
            if isinstance(inp, ITimeDelayAdapter):
                local_time = inp.with_delay(target_time)
                delayed = True

        if not isinstance(inp, NoDependencyAdapter) and not inp.is_static:
            comp = output_owners[inp]
            if not isinstance(comp, ITimeComponent) or (
                isinstance(comp, ITimeComponent) and inp.time < local_time
            ):
                if inp not in deps or local_time > deps[inp][0]:
                    deps[inp] = (local_time, delayed)

    return deps


def _dead_link_error(component, chain, first_index, last_index):
    link_message = ""
    for i, item in enumerate(reversed(chain)):
        link_message += item.name
        if i < len(chain) - 1:
            link_message += (
                " >/> " if i == first_index or i + 1 == last_index else " >> "
            )

    return FinamConnectError(
        f"Dead link detected between "
        f"{chain[0].name} and {str(component)}->{chain[-1].name}:\n"
        f"{link_message}"
    )



"""
Tools for using FINAM.

Logging helper
==============

.. autosummary::
   :toctree: generated

    ErrorLogger
    LogCStdOutStdErr
    LogStdOutStdErr
    LogWriter
    is_loggable
    add_logging_level

Datetime helper
===============

.. autosummary::
   :toctree: generated

    is_timedelta

CWD helper
==========

.. autosummary::
   :toctree: generated

    execute_in_cwd
    set_directory

ENUM helper
===========

.. autosummary::
   :toctree: generated

    get_enum_value

Inspection helper
=================

.. autosummary::
   :toctree: generated

    inspect

Connect helper
==============

.. autosummary::
   :toctree: generated

    ConnectHelper
    FromInput
    FromOutput
    FromValue
"""
from .connect_helper import ConnectHelper, FromInput, FromOutput, FromValue
from .cwd_helper import execute_in_cwd, set_directory
from .date_helper import is_timedelta
from .enum_helper import get_enum_value
from .inspect_helper import inspect
from .log_helper import (
    ErrorLogger,
    LogCStdOutStdErr,
    LogStdOutStdErr,
    LogWriter,
    add_logging_level,
    is_loggable,
)

__all__ = ["execute_in_cwd", "set_directory"]
__all__ += ["is_timedelta"]
__all__ += ["get_enum_value"]
__all__ += ["inspect"]
__all__ += [
    "add_logging_level",
    "ErrorLogger",
    "is_loggable",
    "LogWriter",
    "LogStdOutStdErr",
    "LogCStdOutStdErr",
]
__all__ += ["ConnectHelper", "FromInput", "FromOutput", "FromValue"]



"""
Adapters that deal with time, like temporal interpolation and integration.
"""
import os
from abc import ABC, abstractmethod
from datetime import datetime, timedelta

import numpy as np

from finam.interfaces import NoBranchAdapter, NoDependencyAdapter

from ..data import tools as dtools
from ..errors import FinamNoDataError, FinamTimeError
from ..sdk import Adapter, TimeDelayAdapter
from ..tools.date_helper import is_timedelta
from ..tools.log_helper import ErrorLogger

__all__ = [
    "NextTime",
    "PreviousTime",
    "LinearTime",
    "StepTime",
    "StackTime",
    "DelayFixed",
    "DelayToPush",
    "DelayToPull",
    "TimeCachingAdapter",
    "check_time",
    "interpolate",
    "interpolate_step",
]


class DelayFixed(TimeDelayAdapter):
    """Delays/offsets the request time by subtracting a fixed offset.

    Delayed times that are located before the initial pull/request time are set to this time.

    An illustrative example:
    Component A has a step of 10 days.
    The adapter has a delay of 11 days to guarantee data availability in B.

    .. code-block:: Text

        A  O=========O---------o
        ^           .<---------'
        |           V
        B  =O=O=O=O=O

    Parameters
    ----------

    delay : datetime.timedelta
        The delay duration to subtract from the request time.

    See also
    --------

    .adapters.DelayToPull : Delays to use data from a previous pull.
    .adapters.DelayToPush : Delays to use data from the last push.

    """

    def __init__(self, delay):
        super().__init__()

        with ErrorLogger(self.logger):
            if not is_timedelta(delay):
                raise ValueError("Step must be of type timedelta or relativedelta")

        self.delay = delay

    def with_delay(self, time):
        off = time - self.delay
        if off < self.initial_time:
            return self.initial_time

        return off


# pylint: disable=too-many-ancestors
class DelayToPush(TimeDelayAdapter, NoDependencyAdapter):
    """Delays/offsets the request time to the last push time if out of range.

    An illustrative example:
    The adapter delays time to the last available push date.

    .. code-block:: Text

        A  O=========O---------o
        ^               .<-----'
        |               |
        B  =O=O=O=O=O=O=O

    However, if data for the requested time is available, time is not modified:

    .. code-block:: Text

        A  O=========O---------o
        ^                      ^
        |                      |
        B  =O=O=O=O=O=O=O=O=O=O=O

    If the requested time is before the last push, it is not modified.

    .. note::
        This adapters fully breaks dependency chains and loops.

        It is recommended to use other subclasses of :class:`.ITimeDelayAdapter`,
        e.g. :class:`.adapters.DelayFixed` or :class:`.adapters.DelayToPull`.
        These adapters have a more consistent pull interval, and dependencies are still checked.

    See also
    --------

    .adapters.DelayFixed : Delays to use data with a fixed offset.
    .adapters.DelayToPull : Delays to use data from a previous pull.
    """

    def __init__(self):
        super().__init__()
        self.push_time = None

    def with_delay(self, time):
        if self.push_time is None:
            return self.initial_time

        if time > self.push_time:
            return self.push_time

        return time

    def _source_updated(self, time):
        """Informs the input that a new output is available.

        Parameters
        ----------
        time : datetime
            Simulation time of the notification.
        """
        check_time(self.logger, time)
        self.push_time = time


# pylint: disable=too-many-ancestors
class DelayToPull(TimeDelayAdapter, NoBranchAdapter):
    """Delays/offsets the request time to a previous pull time.

    An illustrative example:
    With ``step=2``, the adapter delays time by two past pulls:

    .. code-block:: Text

        A  O====O====O====O----o
        ^            .<--------'
        |            |
        B  =O=O=O=O=O=O=O

    Delay can be fine-tuned by using ``additional_offset`` (e.d. 2 days):

    .. code-block:: Text

        A  O====O====O====O----o
        ^          .<----------'
        |          |
        B  =O=O=O=O=O=O=O

    Parameters
    ----------

    steps : int, optional
        The number of pulls to delay. Defaults to 1.
    additional_delay : datetime.timedelta
        Additional delay in time units. Defaults to no delay.

    See also
    --------

    .adapters.DelayFixed : Delays to use data with a fixed offset.
    .adapters.DelayToPush : Delays to use data from the last push.

    """

    def __init__(self, steps=1, additional_delay=timedelta(days=0)):
        super().__init__()
        self.steps = steps
        self.additional_delay = additional_delay
        self._pulls = []

    def with_delay(self, time):
        if len(self._pulls) == 0:
            self._pulls.append(self.initial_time)

        t = self._pulls[0]
        off = t - self.additional_delay
        if off < self.initial_time:
            return self.initial_time

        return off

    def _pulled(self, time):
        self._pulls.append(time)
        while len(self._pulls) > self.steps:
            self._pulls.pop(0)

    def _get_data(self, time, target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        d = self.pull_data(time, target)
        return d


class TimeCachingAdapter(Adapter, NoBranchAdapter, ABC):
    """Abstract base class for time handling and caching adapters."""

    def __init__(self):
        super().__init__()
        self.data = []

    @property
    def needs_push(self):
        return True

    def _source_updated(self, time):
        """Informs the input that a new output is available.

        Parameters
        ----------
        time : datetime
            Simulation time of the notification.
        """
        check_time(self.logger, time)

        data = dtools.strip_time(self.pull_data(time, self), self._input_info.grid)
        self.data.append((time, self._pack(data)))

    def _get_data(self, time, _target):
        """Get the output's data-set for the given time.

        Parameters
        ----------
        time : datetime
            simulation time to get the data for.

        Returns
        -------
        array_like
            data-set for the requested time.
        """
        if len(self.data) == 0:
            raise FinamNoDataError(f"No data available in {self.name}")

        check_time(self.logger, time, (self.data[0][0], self.data[-1][0]))

        data = self._interpolate(time)
        self._clear_cached_data(time)
        return data

    def _clear_cached_data(self, time):
        while len(self.data) > 1 and self.data[1][0] <= time:
            d = self.data.pop(0)
            if isinstance(d[1], str):
                os.remove(d[1])
            else:
                self._total_mem -= d[1].nbytes

    @abstractmethod
    def _interpolate(self, time):
        """Interpolate for the given time"""


class NextTime(TimeCachingAdapter):
    """Time interpolation providing the next future value.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.NextTime()
    """

    def _interpolate(self, time):
        if len(self.data) == 1:
            return self._unpack(self.data[0][1])

        for t, data in self.data:
            if time > t:
                continue

            return self._unpack(data)

        raise FinamTimeError(
            f"Time interpolation failed. This should not happen and is probably a bug. "
            f"Time is {time}."
        )


class PreviousTime(TimeCachingAdapter):
    """Time interpolation providing the newest past value.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.PreviousTime()
    """

    def _interpolate(self, time):
        if len(self.data) == 1:
            return self._unpack(self.data[0][1])

        for i, (t, data) in enumerate(self.data):
            if time > t:
                continue
            if time == t:
                return self._unpack(data)

            _, data_prev = self.data[i - 1]
            return self._unpack(data_prev)

        raise FinamTimeError(
            f"Time interpolation failed. This should not happen and is probably a bug. "
            f"Time is {time}."
        )


class StackTime(TimeCachingAdapter):
    """Stacks all incoming data since the last push.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.StackTime()
    """

    def _interpolate(self, time):
        extract = []

        for t, data in self.data:
            if time > t:
                extract.append((t, self._unpack(data)))
                continue

            extract.append((t, self._unpack(data)))
            break

        arr = np.stack([d[1] for d in extract])
        return dtools.prepare(arr, self.info, time_entries=len(extract))


class LinearTime(TimeCachingAdapter):
    """Linear time interpolation.

    .. plot:: api/plots/interpolation-methods.py

        Illustration of interpolation methods.

    See also
    --------

    .adapters.StepTime : Step-wise time interpolation.
    .adapters.AvgOverTime : Average aggregation over time.
    .adapters.SumOverTime : Sum aggregation over time.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.LinearTime()
    """

    def _interpolate(self, time):
        if len(self.data) == 1:
            return self.data[0][1]

        for i, (t, data) in enumerate(self.data):
            if time > t:
                continue
            if time == t:
                return self._unpack(data)

            t_prev, data_prev = self.data[i - 1]

            dt = (time - t_prev) / (t - t_prev)

            result = interpolate(self._unpack(data_prev), self._unpack(data), dt)

            return result

        raise FinamTimeError(
            f"Time interpolation failed. This should not happen and is probably a bug. "
            f"Time is {time}."
        )


class StepTime(TimeCachingAdapter):
    """Step-wise time interpolation.

    .. plot:: api/plots/interpolation-methods.py

        Illustration of interpolation methods.

    Parameters
    ----------

    step : float
        Value in range [0, 1] that determines the relative step position.
        For a value of 0.0, the new value is returned for any dt > 0.0.
        For a value of 1.0, the old value is returned for any dt <= 1.0.
        Values between 0.0 and 1.0 shift the step between the first and the second time.
        A value of 0.5 results in nearest interpolation.

    See also
    --------

    .adapters.LinearTime : Linear time interpolation.
    .adapters.AvgOverTime : Average aggregation over time.
    .adapters.SumOverTime : Sum aggregation over time.

    Examples
    --------

    .. testcode:: constructor

        import finam as fm

        adapter = fm.adapters.StepTime(step=0.0)
    """

    def __init__(self, step=0.5):
        super().__init__()
        self.step = step

    def _interpolate(self, time):
        if len(self.data) == 1:
            return self.data[0][1]

        for i, (t, data) in enumerate(self.data):
            if time > t:
                continue
            if time == t:
                return self._unpack(data)

            t_prev, data_prev = self.data[i - 1]

            dt = (time - t_prev) / (t - t_prev)

            result = interpolate_step(data_prev, data, dt, self.step)

            return self._unpack(result)

        raise FinamTimeError(
            f"Time interpolation failed. This should not happen and is probably a bug. "
            f"Time is {time}."
        )


def interpolate(old_value, new_value, dt):
    """Interpolate between old and new value.

    Parameters
    ----------
    old_value : array_like
        Old value.
    new_value : array_like
        New value.
    dt : float
        Time step between values.

    Returns
    -------
    array_like
        Interpolated value.
    """
    return old_value + dt * (new_value - old_value)


def interpolate_step(old_value, new_value, dt, step):
    """Interpolate step-wise between old and new value.

    Parameters
    ----------
    old_value : array_like
        Old value.
    new_value : array_like
        New value.
    dt : float
        Time step between values.
    step : float
        Value in range [0, 1] that determines the relative step position.

        * For a value of 0.0, the new value is returned for any dt > 0.0.
        * For a value of 1.0, the old value is returned for any dt <= 1.0.
        * Values between 0.0 and 1.0 shift the step between the first and the second time.
        * A value of 0.5 results in nearest interpolation.

    Returns
    -------
    array_like
        Interpolated value.
    """
    return new_value if dt > step else old_value


def check_time(logger, time, time_range=(None, None)):
    """
    Checks time.

    Checks time for being of type :class:`datetime <datetime.datetime>`, and to be in range of time_range
    (upper and lower limits inclusive).

    Parameters
    ----------
    logger : logging.Logger
        Logger to print to
    time : any
        Time to be tested
    time_range : tuple, optional
        Tuple of (min, max) time, elements can be `None`, by default (None, None)

    Raises
    ------
    FinamTimeError
        if any of the checks fails
    """
    with ErrorLogger(logger):
        if not isinstance(time, datetime):
            raise FinamTimeError("Time must be of type datetime")

        if time_range[1] is not None and time > time_range[1]:
            raise FinamTimeError(
                "Requested data for time point in the future. "
                f"Latest data: {time_range[1]}, request: {time}"
            )

        if time_range[0] is not None and time < time_range[0]:
            raise FinamTimeError(
                "Requested data for time point in the past. "
                f"Earliest data: {time_range[0]}, request: {time}"
            )
